#!/usr/bin/env python3
import argparse
from contextlib import contextmanager


@contextmanager
def subparser(parser, name, description):
    yield parser.add_parser(name, conflict_handler='resolve', help=description)

@contextmanager
def mutex_group(parser, required):
    group = parser.add_argument_group(f'Mutually exclusive '
                                      f'{"required" if required else "optional"} '
                                      f'arguments')
    yield group.add_mutually_exclusive_group(required=required)

@contextmanager
def arg_group(parser, name):
    yield parser.add_argument_group(name)

def __all_plots(group, required=False):
    group.add_argument('--all_plots', action='store_true', required=required,
                       help='Generate a combined plot for CUGO, AA sequence length and TMHMM')

def __all_proteins(group, required=False):
    group.add_argument('-a', '--all_proteins', type=str, required=required,
                       help='Path to FASTA file containing all GlobDB protein sequences')

def __all_metadata(group, required=False):
    group.add_argument('--all_metadata', action='store_true', required=required,
                       help='Include all metadata in output file')

def __annotation(group, required=False):
    group.add_argument('--annotation', type=str, default='COG_ID', required=required,
                       help='Annotation: COG_ID, KEGG_ID, Pfam_ID (default: COG_ID)')

def __bin_width(group, required=False):
    group.add_argument('-b', '--bin_width', type=int, default=50, required=required,
                       help='Bin width for amino acid sequence size plotting (default: 50)')

def __blast_output(group, required=False):
    group.add_argument('-b', '--blast_output', type=str, default=None, required=required,
                       help='Path to BLAST/DIAMOND output file')

def __block(group, required=False):
    group.add_argument('-b', '--block', type=int, default=6, required=required,
                       help='blastp sequence block size in billions of letters (default: 6)')

def __bsr(group, required=False):
    group.add_argument('-b', '--bsr', type=str, required=required,
                       help='Path to blast score ratio input file.')

def __bsr_cutoff(group, required=False):
    group.add_argument('-b', '--bsr_cutoff', type=float, default=None, required=required,
                       help="Blast Score Ratio cutoff for inclusion in updated dataset")

def __chunk(group, required=False):
    group.add_argument('-c', '--chunk', type=int, default=2, required=required,
                       help='Number of chunks for diamond blastp index processing (default: 2)')

def __cluster_path(group, required=False):
    group.add_argument('-c', '--cluster_path', type=str, default=None, required=required,
                       help='Path to cluster TSV file')

def __create_yaml(group, required=False):
    group.add_argument('--create_yaml', action='store_true', required=required,
                       help='Create metadata yaml file from command line parameters')

def __cog_gff(group, required=False):
    group.add_argument('-c', '--cog_gff', type=str, required=required,
                       help='Path to (.tar.gz) GFF directory containing COG annotations')

def __column_info_path(group, required=False):
    group.add_argument('-c', '--column_info_path', type=str, default=None, required=required,
                       help='Path to json file containing columns the blast output file generated by aastk search')

def __context_path(group, required=False):
    group.add_argument('-c', '--context_path', type=str, required=required,
                       help='Path to CUGO context file')

def __cugo(group, required=False):
    group.add_argument('--cugo', action='store_true', required=required,
                       help='Generate CUGO plot')

def __cugo_range(group, required=False):
    group.add_argument('-r', '--cugo_range', type=int, required=required,
                       help='CUGO range for genomic context analysis')

def __culture_collection_path(group, required=False):
    group.add_argument('-c', '--culture_collection_path', type=str, required=required,
                       help='Path to culture collection data TSV file')

def __dataset(group, required=False):
    group.add_argument('-d', '--dataset', type=str, required=required,
                       help='Dataset name')

def __diamond_db(group, required=False):
    group.add_argument('-d', '--diamond_db', type=str, required=required,
                       help='Path to DIAMOND database')

def __dbmin(group, required=False):
    group.add_argument('-d', '--dbmin', type=int, default=None, required=required,
                       help='Lower database score cutoff for inclusion in updated dataset')

def __db_path(group, required=False):
    group.add_argument('-d', '--db_path', type=str, default='temp_genome_data.db', required=required,
                       help='Path to SQLite database')

def __exaggeration(group, required=False):
    group.add_argument('-e', '--exaggeration', type=int, default=6, required=required,
                       help='Exaggeration value for tSNE embedding (default: 6)')

def __extracted(group, required=False):
    group.add_argument('-e', '--extracted', type=str, required=required,
                       help='Path to FASTA file containing extracted matching sequences')

def __fasta(group, required=False):
    group.add_argument('-f', '--fasta', type=str, required=required,
                       help='Path to FASTA file')

def __filter_seqs(group, required=False):
    group.add_argument('--filter_seqs', action='store_true', required=required,
                       help='Filter datasets in FASTA format for improved homogeneity')

def __force(group, required=False):
    group.add_argument('--force', action='store_true', required=required,
                       help='Overwrite existing files in specified path')

def __flank_lower(group, required=False):
    group.add_argument('-l', '--flank_lower', type=int, required=required,
                       help='Start of flanking window (inclusive)')

def __flank_upper(group, required=False):
    group.add_argument('-u', '--flank_upper', type=int, required=required,
                       help='End of flanking window (inclusive)')

def __globdb_version(group, required=False):
    group.add_argument('-g', '--globdb_version', type=str, required=required,
                       help='GlobDB version (example: r226)')

def __help(group, required=False):
    group.add_argument('-h', '--help', action='help',
                       help='Display help text')

def __high_level_environment_path(group, required=False):
    group.add_argument('-h', '--high_level_environment_path', type=str, required=required,
                       help='Path to high level environment data TSV file')

def __id_list(group, required=False):
    group.add_argument('-i', '--id_list', type=str, required=required,
                       help='Path to list of GlobDB protein IDs')

def __include_annotation(group, required=False):
    group.add_argument('--include_annotation', action='store_true', required=required,
                       help='Include annotation metadata in output')

def __include_culture_collection(group, required=False):
    group.add_argument('--include_culture_collection', action='store_true', required=required,
                       help='Include culture collection metadata in output')

def __include_high_level_environment(group, required=False):
    group.add_argument('--include_high_level_environment', action='store_true', required=required,
                       help='Include high level environment metadata in output')

def __include_low_level_environment(group, required=False):
    group.add_argument('--include_low_level_environment', action='store_true', required=required,
                       help='Include low level environment metadata in output')

def __include_taxonomy(group, required=False):
    group.add_argument('--include_taxonomy', action='store_true', required=required,
                       help='Include taxonomy metadata in output')

def __iterations(group, required=False):
    group.add_argument('-i', '--iterations', type=str, default=500, required=required,
                       help='Number of tSNE embedding iterations (default: 500)')

def __keep(group, required=False):
    group.add_argument('--keep', action='store_true', required=required,
                       help='Keep intermediate files in specified paths')

def __kegg_gff(group, required=False):
    group.add_argument('-k', '--kegg_gff', type=str, required=required,
                       help='Path to (.tar.gz) GFF directory containing KEGG annotations')

def __key_column(group, required=False):
    group.add_argument('-k', '--key_column', type=int, default=0, required=required,
                       help='Column index in the BLAST tab file to pull unique IDs from (default: 0)')

def __low_level_environment_path(group, required=False):
    group.add_argument('-l', '--low_level_environment_path', type=str, required=required,
                       help='Path to high level environment data TSV file')

def __matched(group, required=False):
    group.add_argument('-m', '--matched', type=str, required=required,
                       help='Path to matched sequences FASTA file from previous PASR run')

def __matrix(group, required=False):
    group.add_argument('-m', '--matrix', type=str, default='BLOSUM45', required=required,
                       help='BLOSUM substitution matrix: BLOSUM45, BLOSUM62 (default: BLOSUM45)')

def __matrix_path(group, required=False):
    group.add_argument('-m', '--matrix_path', type=str, required=required,
                       help='Path to alignment matrix .npy file')

def __max_score_max(group, required=False):
    group.add_argument('-u', '--max_score_max', type=int, required=required,
                       help='Upper cutoff for max score range')

def __max_score_min(group, required=False):
    group.add_argument('-l', '--max_score_min', type=int, required=required,
                       help='Lower cutoff for max score range')

def __max_scores(group, required=False):
    group.add_argument('-m', '--max_scores', type=str, required=required,
                       help='Path to file containing max scores')

def __metadata_protein(group, required=False):
    group.add_argument('--metadata_protein', type=str, required=required,
                       help='Select metadata for plotting; run "aastk metadata_categories" to view available categories"')

def __metadata_matrix(group, required=False):
    group.add_argument('--metadata_matrix', type=str, required=required,
                       help='Path to alignment matrix metadata file')

def __no_cluster(group, required=False):
    group.add_argument('-n', '--no_cluster', type=int, required=required,
                       help='Number of cluster of choice in TSV file')

def __output(group, required=False):
    group.add_argument('-o', '--output', type=str, required=required,
                       help='Output directory (default: current working directory)')

def __params(group, required=False):
    group.add_argument('--params', action='store_true', required=required,
                       help='Select subset of matched sequences based on command line parameters')

def __perplexity(group, required=False):
    group.add_argument('-p', '--perplexity', type=int, default=50, required=required,
                       help='Perplexity value for tSNE embedding (default: 50)')

def __pfam_gff(group, required=False):
    group.add_argument('-p', '--pfam_gff', type=str, required=required,
                       help='Path to (.tar.gz) GFF directory containing Pfam annotations')

def __position(group, required=False):
    group.add_argument('-p', '--position', type=int, required=required,
                       help='CUGO position to retrieve protein IDs for')

def __protein_ids(group, required=False):
    group.add_argument('-p', '--protein_ids', type=str, required=required,
                       help='Path to file containing list of protein IDs')

def __query(group, required=False):
    group.add_argument('-q', '--query', type=str, default=None, required=required,
                       help='Path to DIAMOND query FASTA file')

def __score_column(group, required=False):
    group.add_argument('-s', '--score_column', type=int, default=None, required=required,
                       help='Index of the raw score column of BLAST output file')

def __seed(group, required=False):
    group.add_argument('-s', '--seed', type=str, default=None, required=required,
                       help='Path to seed FASTA file for DIAMOND database creation')

def __sensitivity(group, required=False):
    group.add_argument('--sensitivity', required=required,
                       help='DIAMOND search sensitivity level: fast, sensitive, mid-sensitive, very-sensitive, '
                            'ultra-sensitive, or faster (default: fast)')

def __show(group, required=False):
    group.add_argument('-s', '--show', action='store_true', required=required,
                       help='Show cluster number at cluster centers in output plots')

def __size(group, required=False):
    group.add_argument('--size', action='store_true', required=required,
                       help='Generate AA sequence length plot')

def __sql(group, required=False):
    group.add_argument('--sql', action='store_true', required=required,
                       help='Retrieve sequences from AASTK SQLite database')

def __subset(group, required=False):
    group.add_argument('-s', '--subset', type=str, required=required,
                       help='Path to subset FASTA')

def __subset_size(group, required=False):
    group.add_argument('--subset_size', type=int, required=required,
                       help='Number of sequences in subset FASTA file')

def __svg(group, required=False):
    group.add_argument('--svg', action='store_true', required=required,
                       help='Generate plot in SVG format')

def __taxonomy_path(group, required=False):
    group.add_argument('-t', '--taxonomy_path', type=str, default=None, required=required,
                       help='Path to gzipped taxonomy TSV file')

def __threads(group, required=False):
    group.add_argument('-n', '--threads', type=int, default=1, required=required,
                       help='Number of threads (default: 1)')

def __tmhmm_dir(group, required=False):
    group.add_argument('-t', '--tmhmm_dir', type=str, required=required,
                       help='Directory containing tmhmm files')

def __tmh_y_range(group, required=False):
    group.add_argument('-t', '--tmh_y_range', type=int, required=required,
                       help='Upper limit for TMH plot y-axis (default: maximum number of transmembrane helices)')

def __top_n(group, required=False):
    group.add_argument('-t', '--top_n', type=int, default=3, required=required,
                       help='Number of top annotations to plot per position (default: 3)')

def __tsv(group, required=False):
    group.add_argument('-t', '--tsv', type=str, required=required,
                       help='Path to ASM_clust tab separated output file')

def __update(group, required=False):
    group.add_argument('--update', action='store_true', required=required,
                       help='Update subset of data using metadata yaml file')

def __yaml(group, required=False):
    group.add_argument('-y', '--yaml', type=str, required=required,
                       help='Path to metadata yaml file')

def __y_range(group, required=False):
    group.add_argument('-y', '--y_range', type=int, required=required,
                       help='Upper limit for sequence length plot y-axis (default: maximum length)')

def get_main_parser():
    main_parser = argparse.ArgumentParser(
        prog='aastk', add_help=False, conflict_handler='resolve')

    main_parser.add_argument('--silent', action='store_true',
                             help='Suppress all console output except errors')
    sub_parsers = main_parser.add_subparsers(help="--", dest='subparser_name')


    ### PARSER FOR PASR FUNCTIONALITIES AND WORKFLOW ###
    with subparser(sub_parsers, 'build', 'Build DIAMOND database from seed sequence(s)') as parser:
        with arg_group(parser, 'Required arguments') as grp:
            __seed(grp, required=True)
        with arg_group(parser, 'Optional') as grp:
            __output(grp)
            __threads(grp)
            __force(grp)

    with subparser(sub_parsers, 'search', 'Search DIAMOND reference database for homologous sequences') as parser:
        with arg_group(parser, 'Required arguments') as grp:
            __diamond_db(grp, required=True),
            __query(grp, required=True),
        with arg_group(parser, 'Optional') as grp:
            __output(grp)
            __threads(grp)
            __block(grp)
            __chunk(grp)
            __sensitivity(grp)
            __force(grp)

    with subparser(sub_parsers, 'get_hit_seqs', 'Extract reads that have DIAMOND hits against custom database') as parser:
        with mutex_group(parser, required=True) as grp:
            __sql(grp)
            __query(grp)
        with arg_group(parser, 'Required arguments') as grp:
            __blast_output(grp, required=True)
        with arg_group(parser, 'Optional') as grp:
            __output(grp)
            __db_path(grp)
            __threads(grp)
            __key_column(grp)
            __force(grp)

    with subparser(sub_parsers, 'max_score', 'Calculate max scores for extracted sequences using BLOSUM matrix') as parser:
        with arg_group(parser, 'Required arguments') as grp:
            __extracted(grp, required=True)
        with arg_group(parser, 'Optional') as grp:
            __output(grp)
            __matrix(grp)
            __force(grp)

    with subparser(sub_parsers, 'bsr', 'Compute BSR (Blast Score Ratio) using a BLAST tab file and max scores from a TSV.') as parser:
        with mutex_group(parser, required=True) as grp:
            __column_info_path(grp)
            __score_column(grp)
        with arg_group(parser, 'Required arguments') as grp:
            __blast_output(grp, required=True)
            __max_scores(grp, required=True)
        with arg_group(parser, 'Optional') as grp:
            __output(grp)
            __key_column(grp)
            __force(grp)

    with subparser(sub_parsers, 'pasr_plot', 'Plot the Blast Score Ratio of query sequences against the DIAMOND database') as parser:
        with arg_group(parser, 'Required arguments') as grp:
            __bsr(grp, required=True)
        with arg_group(parser, 'Optional') as grp:
            __output(grp)
            __svg(grp)
            __force(grp)
            __update(grp)
            __yaml(grp)


    with subparser(sub_parsers, 'pasr_select', 'Select target sequences in accordance with metadata cutoffs') as parser:
        with mutex_group(parser, required=True) as grp:
            __yaml(grp)
            __params(grp)
        with arg_group(parser, 'Required arguments') as grp:
            __matched(grp, required=True)
            __bsr(grp, required=True)
        with arg_group(parser, 'Optional') as grp:
            __output(grp)
            __create_yaml(grp)
            __max_score_min(grp)
            __max_score_max(grp)
            __dbmin(grp)
            __bsr_cutoff(grp)
            __force(grp)


    with subparser(sub_parsers, 'pasr', 'PASR: protein alignment score ratio') as parser:
        with arg_group(parser, 'Required arguments') as grp:
            __query(grp, required=True)
            __seed(grp, required=True)
        with arg_group(parser, 'Optional') as grp:
            __output(grp)
            __matrix(grp)
            __threads(grp)
            __db_path(grp)
            __key_column(grp)
            __block(grp)
            __chunk(grp)
            __sensitivity(grp)
            __update(grp)
            __yaml(grp)
            __sql(grp)
            __keep(grp)
            __svg(grp)
            __force(grp)

    ### PARSER FOR CUGO FUNCTIONALITIES AND WORKFLOW ###
    with subparser(sub_parsers, 'database', 'Create AASTK SQLite database') as parser:
        with arg_group(parser, 'Required arguments') as grp:
            __all_proteins(grp, required=True)
        with arg_group(parser, 'Optional') as grp:
            __cog_gff(grp)
            __kegg_gff(grp)
            __pfam_gff(grp)
            __taxonomy_path(grp)
            __culture_collection_path(grp)
            __high_level_environment_path(grp)
            __low_level_environment_path(grp)
            __globdb_version(grp)
            __output(grp)
            __tmhmm_dir(grp)

    with subparser(sub_parsers, 'database_check', 'Checks for missing data in AASTK SQLite database') as parser:
        with arg_group(parser, 'Required arguments') as grp:
            __db_path(grp, required=True)
        with arg_group(parser, 'Optional') as grp:
            __output(grp)
            __force(grp)

    with subparser(sub_parsers, 'meta', 'Retrieve metadata from AASTK SQLite database') as parser:
        with arg_group(parser, 'Required arguments') as grp:
            __db_path(grp, required=True)
            __fasta(grp, required=True)
        with arg_group(parser, 'Optional') as grp:
            __output(grp)
            __threads(grp)
            __include_annotation(grp)
            __include_culture_collection(grp)
            __include_high_level_environment(grp)
            __include_low_level_environment(grp)
            __include_taxonomy(grp)
            __all_metadata(grp)
            __force(grp)




    with subparser(sub_parsers, 'context', 'Parse context information from CUGO input file') as parser:
        with mutex_group(parser, required=True) as grp:
            __fasta(grp)
            __id_list(grp)
        with arg_group(parser, 'Required arguments') as grp:
            __db_path(grp, required=True)
            __cugo_range(grp, required=True)
        with arg_group(parser, 'Optional') as grp:
            __annotation(grp)
            __output(grp)
            __threads(grp)
            __force(grp)

    with subparser(sub_parsers, 'cugo_plot', 'Plot CUGO context') as parser:
        with arg_group(parser, 'Required arguments') as grp:
            __context_path(grp, required=True)
            __flank_lower(grp, required=True)
            __flank_upper(grp, required=True)
        with arg_group(parser, 'Optional') as grp:
            __annotation(grp)
            __top_n(grp)
            __cugo(grp)
            __size(grp)
            __all_plots(grp)
            __bin_width(grp)
            __y_range(grp)
            __tmh_y_range(grp)
            __output(grp)
            __force(grp)
            __svg(grp)



    with subparser(sub_parsers, 'cugo', 'CUGO: Co-localized gene organization') as parser:
        with mutex_group(parser, required=True) as grp:
            __fasta(grp)
            __id_list(grp)
        with arg_group(parser, 'Required arguments') as grp:
            __db_path(grp, required=True)
            __cugo_range(grp, required=True)
            __flank_lower(grp, required=True)
            __flank_upper(grp, required=True)
        with arg_group(parser, 'Optional') as grp:
            __annotation(grp)
            __output(grp)
            __top_n(grp)
            __bin_width(grp)
            __y_range(grp)
            __tmh_y_range(grp)
            __output(grp)
            __threads(grp)
            __svg(grp)
            __force(grp)


    with subparser(sub_parsers, 'cugo_select', 'Retrieve protein IDs for select CUGO position') as parser:
        with arg_group(parser, 'Required arguments') as grp:
            __context_path(grp, required=True)
            __position(grp, required=True)
            __db_path(grp, required=True)
        with arg_group(parser, 'Optional') as grp:
            __output(grp)
            __threads(grp)
            __filter_seqs(grp)
            __force(grp)

    with subparser(sub_parsers, 'filter', 'Filter datasets in FASTA format for improved homogeneity') as parser:
        with arg_group(parser, 'Required arguments') as grp:
            __fasta(grp, required=True)
        with arg_group(parser, 'Optional') as grp:
            __output(grp)
            __threads(grp)
            __sql(grp)
            __db_path(grp)
            __force(grp)

    with subparser(sub_parsers, 'matrix', 'Create alignment matrix for tSNE embedding and DBSCAN clustering') as parser:
        with mutex_group(parser, required=True) as grp:
            __subset(grp)
            __subset_size(grp)
        with arg_group(parser, 'Required arguments') as grp:
            __fasta(grp, required=True)
        with arg_group(parser, 'Optional') as grp:
            __output(grp)
            __threads(grp)
            __force(grp)

    with subparser(sub_parsers, 'cluster', 'Run tSNE embedding and DBSCAN clustering on input matrix and matrix metadata') as parser:
        with arg_group(parser, 'Required arguments') as grp:
            __matrix_path(grp, required=True)
            __metadata_matrix(grp, required=True)
        with arg_group(parser, 'Optional') as grp:
            __output(grp)
            __perplexity(grp)
            __iterations(grp)
            __exaggeration(grp)
            __threads(grp)
            __force(grp)


    with subparser(sub_parsers, 'casm_plot', 'Plot CASM .tsv output files') as parser:
        with arg_group(parser, 'Required arguments') as grp:
            __cluster_path(grp, required=True)
        with arg_group(parser, 'Optional') as grp:
            __output(grp)
            __db_path(grp)
            __metadata_protein(grp)
            __show(grp)
            __svg(grp)
            __force(grp)

    with subparser(sub_parsers, 'casm', 'CASM: protein clustering using alignment score matrices') as parser:
        with mutex_group(parser, required=True) as grp:
            __subset(grp)
            __subset_size(grp)
        with arg_group(parser, 'Required arguments') as grp:
            __fasta(grp, required=True)
        with arg_group(parser, 'Optional') as grp:
            __threads(grp)
            __output(grp)
            __db_path(grp)
            __perplexity(grp)
            __iterations(grp)
            __exaggeration(grp)
            __metadata_protein(grp)
            __force(grp)
            __keep(grp)
            __show(grp)
            __svg(grp)

    with subparser(sub_parsers, 'casm_select', 'Pick CASM clusters to generate .faa file for further analysis') as parser:
        with arg_group(parser, 'Required arguments') as grp:
            __cluster_path(grp, required=True)
            __fasta(grp, required=True)
            __no_cluster(grp, required=True)
        with arg_group(parser, 'Optional') as grp:
            __output(grp)
            __force(grp)

    with subparser(sub_parsers, 'list_metadata', 'Display avaiilable metadata categories') as parser:
        pass

    with subparser(sub_parsers, 'export_fasta', 'Create FASTA file containing all GlobDB protein sequences') as parser:
        with arg_group(parser, 'Required arguments') as grp:
            __db_path(grp, required=True)
        with arg_group(parser, 'Optional') as grp:
            __output(grp)
            __threads(grp)
            __force(grp)


    return main_parser



#!/usr/bin/env python3
from .util import *

import subprocess
import matplotlib.pyplot as plt
import pandas as pd
import yaml
import logging
import json
from pathlib import Path
import numpy as np
from mpl_toolkits.axes_grid1.inset_locator import inset_axes
import sqlite3
from tqdm import tqdm
from concurrent.futures import ThreadPoolExecutor, as_completed

logger = logging.getLogger(__name__)

BLOSUM_DIAGONALS = {
    "BLOSUM45": {
        'A': 5, 'R': 7, 'N': 6, 'D': 7, 'C': 12, 'Q': 6, 'E': 6, 'G': 7,
        'H': 10, 'I': 5, 'L': 5, 'K': 5, 'M': 6, 'F': 8, 'P': 9, 'S': 4,
        'T': 5, 'W': 15, 'Y': 8, 'V': 5, 'B': 5, 'J': 4, 'Z': 5, 'X': 0
    },
    "BLOSUM62": {
        'A': 4, 'R': 5, 'N': 6, 'D': 6, 'C': 9, 'Q': 5, 'E': 5, 'G': 6,
        'H': 8, 'I': 4, 'L': 4, 'K': 5, 'M': 5, 'F': 6, 'P': 7, 'S': 4,
        'T': 5, 'W': 11, 'Y': 7, 'V': 4, 'B': 4, 'J': 3, 'Z': 4, 'X': 0
    }
}

# ===============================
# aastk build CLI FUNCTION
# ===============================
def build(seed_fasta: str,
          threads: int,
          output: str,
          force: bool = False):
    """
    Builds a DIAMOND protein database from a seed FASTA file.

    Args:
        seed_fasta (str): Path to the FASTA file containing seed sequences.
        threads (int): Number of threads to use.
        output (str): Directory where the database should be stored. (Default: current working directory)
        force (bool): If true, existing files/directories in output path are overwritten

    Returns:
        db_path: Path to DIAMOND protein database.

    Raises:
        RuntimeError: If the DIAMOND database creation fails.
    """
    # check if diamond is in path
    check_dependency_availability('diamond')

    if Path(seed_fasta).is_file():
        pass
    else:
        raise FileNotFoundError(f"Seed FASTA file does not exist: {seed_fasta}")

    seed_fasta_filename = Path(seed_fasta).name
    protein_name = determine_dataset_name(seed_fasta_filename, '.', 0)

    # ===============================
    # Database path setup
    # ===============================
    db_path = ensure_path(path=output, target=f"{protein_name}_seed_db", suffix='.dmnd', force=force)

    # log the path
    logger.info(f"Building DIAMOND database for {protein_name} at {db_path}")

    # =======================================
    # DIAMOND makedb command construction
    # =======================================
    cmd = ["diamond", "makedb",
           "--in", seed_fasta,
           "-d", db_path,
           "-p", str(threads)]
    logger.info(f"Running command: {' '.join(cmd)}")

    # ===============================
    # Execute DIAMOND makedb
    # ===============================
    try:
        proc = subprocess.Popen(
            cmd,
            stdout=subprocess.DEVNULL,
            stderr=subprocess.PIPE,
            text=True
        )
        _, stderr = proc.communicate()

        if proc.returncode != 0:
            logger.error(f"DIAMOND makedb failed with return code {proc.returncode}")
            if stderr:
                logger.error(f"STDERR: {stderr}")
            raise RuntimeError(f"DIAMOND database creation failed with return code {proc.returncode}")

        if stderr:
            logger.log(99, stderr)

    except Exception as e:
        if not isinstance(e, RuntimeError):
            logger.error(f"Unexpected error in building the DIAMOND database: {e}")
            raise RuntimeError(f"DIAMOND database creation failed: {e}") from e
        raise

    db_file = Path(f"{db_path}.dmnd")
    if not db_file.exists():
        raise RuntimeError(f"DIAMOND database was not created at {db_file}")


    logger.info(f"Successfully built DIAMOND database at {db_file}")
    return db_path





# ===============================
# aastk search CLI FUNCTION
# ===============================
def search(db_path: str,
           query_path: str,
           threads: int,
           output_dir: str,
           sensitivity: str,
           block: int = 6,
           chunk: int = 2,
           force: bool = False):
    """
    Searches a DIAMOND reference database for homologous sequences.

    Args:
        db_path (str): Path to the DIAMOND reference database.
        query_path (str): Path to the query FASTA file containing sequences to be searched.
        threads (int): Number of CPU threads to use for the search.
        output_dir (str): Directory where results should be stored. (Default: current working directory)
        sensitivity (str): Choose sensitivity of diamond blastp search (Default: --fast)
        block (int): Choose diamond blastp sequence block size in billions of letters. (Default: 6)
        chunk (int): Choose number of chunks for diamond blastp index processing. (Default: 2)
        force (bool): If true, existing files/directories in output path are overwritten


    Returns:
        output_path: Path to tabular BLAST output file.
    """
    # check if diamond is in PATH
    check_dependency_availability('diamond')

    # automatically name files
    protein_name = determine_dataset_name(db_path, '.', 0, '_seed_db')

    # ===============================
    # Output file path setup
    # ===============================
    output_path = ensure_path(output_dir, f"{protein_name}_hits.txt", force=force)
    column_info_path = ensure_path(output_dir, f"{protein_name}_columns.json", force=force)

    # =======================================================
    # DIAMOND blastp output file column configuration
    # =======================================================
    # define the output columns of interest
    columns = ["qseqid", "sseqid", "pident", "qlen", "slen", "length", "mismatch", "gapopen", "qstart", "qend",
               "sstart", "send", "evalue", "bitscore", "score"]

    # save column information to json file as to not hardcode the score column in the bsr function
    column_info = {col: idx for idx, col in enumerate(columns)}
    with open(column_info_path, 'w') as f:
        json.dump(column_info, f, indent=2)

    logger.info(f"Saved column information to {column_info_path}")

    # ===============================
    # Parameter setup
    # ===============================
    # check for sensitivity, if None set to default --fast
    sensitivity_param = f"--{sensitivity}" if sensitivity else "--fast"

    logger.info(f"Searching DIAMOND database {db_path} with query {query_path}")
    logger.info(f"Output path: {output_path}")
    logger.info(f"Using parameters: sensitivity={sensitivity_param}, block={block}, chunk={chunk}")

    # =======================================
    # DIAMOND blastp command construction
    # =======================================
    cmd = ["diamond", "blastp",
           "-d", db_path,
           "-q", query_path,
           "-p", str(threads),
           "-o", output_path,
           "-k", str(1),
           "--matrix", "blosum45",
           "--masking", str(0),
           "--outfmt", str(6), *columns,
           "-b", str(block),
           "-c", str(chunk),
           "--min-score", str(50),
           "--comp-based-stats", str(0),
           sensitivity_param]

    logger.debug(f"Running command: {' '.join(cmd)}")
    # =======================================
    # Execute DIAMOND blastp search
    # =======================================
    try:
        proc = subprocess.Popen(
            cmd,
            stdout=subprocess.DEVNULL,
            stderr=subprocess.PIPE,
            text=True
        )
        _, stderr = proc.communicate()

        if proc.returncode != 0:
            logger.error(f"DIAMOND blastp failed with return code {proc.returncode}")
            if stderr:
                logger.error(f"STDERR: {stderr}")
            raise RuntimeError(f"DIAMOND blastp search failed with return code {proc.returncode}")

        if stderr:
            logger.log(99, stderr)

    except Exception as e:
        if not isinstance(e, RuntimeError):
            logger.error(f"Unexpected error in DIAMOND blastp search: {e}")
            raise RuntimeError(f"DIAMOND blastp search failed: {e}") from e
        raise

    if not Path(output_path).exists():
        logger.error(f"DIAMOND output file not found at {output_path}")
        raise RuntimeError(f"DIAMOND search did not produce output at {output_path}")

    logger.info(f"Successfully completed DIAMOND search. Results at {output_path}")
    return output_path, column_info_path





# ===============================
# aastk get_hit_seqs CLI FUNCTION
# ===============================
def fetch_sequence_batch(db_path: str, batch: list, batch_num: int):
    """
    Fetch a single batch of sequences from the database.

    Args:
        db_path: Path to SQLite database
        batch: List of sequence IDs to fetch
        batch_num: Batch number for tracking

    Returns:
        List of tuples (seqid, sequence)
    """
    conn = sqlite3.connect(db_path)

    placeholders = ','.join('?' * len(batch))

    cursor = conn.execute(f"""
        SELECT seqID, protein_seq 
        FROM protein_data 
        WHERE seqID IN ({placeholders}) AND protein_seq IS NOT NULL
    """, batch)

    results = []
    for seqid, compressed_seq in cursor:
        sequence = decompress_sequence(compressed_seq)
        results.append((seqid, sequence))

    conn.close()
    return results


def get_hit_seqs(blast_tab: str,
                 query_path: str,
                 output_dir: str,
                 db_path: str,
                 key_column: int = 0,
                 threads: int = 1,
                 sql: bool = False,
                 force: bool = False):
    """
    Extracts reads that have BLAST/DIAMOND hits against a custom database.

    Args:
        blast_tab: Tabular BLAST/DIAMOND output file.
        query_path: Fasta or fastq file containing sequencing reads used as BLAST/DIAMOND queries.
        output_dir (str): Directory where extracted sequences should be stored.
        db_path (str): Path to AASTK SQLiteDB
        key_column: Column index in the BLAST tab file to pull unique IDs from (default is 0).
        sql (bool): If true, retrieve sequences from AASTK SQLite database.
        force (bool): If true, existing files/directories in output path are overwritten

    Returns:
        out_fasta: Path to output FASTA file.
    """
    # check if seqkit is in path
    check_dependency_availability('seqkit')

    protein_name = determine_dataset_name(blast_tab, '.', 0, '_hits')

    # ===============================
    # Output file setup
    # ===============================
    out_fasta = ensure_path(output_dir, f"{protein_name}_matched.faa", force=force)
    stats_path = ensure_path(output_dir, f"{protein_name}_matched.stats", force=force)

    # =========================================
    # Extract matching IDs from BLAST results
    # =========================================
    matching_ids = extract_unique_keys(blast_tab, key_column)
    logger.info(f"Extracted {len(matching_ids)} unique matching sequence IDs")

    # ======================================
    # Extract and write matching sequences
    # ======================================
    sequences_written = 0
    batch_size = 900

    with open(out_fasta, "w") as out:
        if sql:
            # convert set to list for batching
            matching_ids_list = list(matching_ids)
            total_ids = len(matching_ids_list)

            # create batches
            batches = [matching_ids_list[i:i + batch_size]
                       for i in range(0, total_ids, batch_size)]

            logger.info(f"Processing {len(batches)} batches with {threads} workers")

            # process batches in parallel
            with ThreadPoolExecutor(max_workers=threads) as executor:
                future_to_batch = {
                    executor.submit(fetch_sequence_batch, db_path, batch, i): i
                    for i, batch in enumerate(batches)
                }

                # process completed batches as they finish
                for future in tqdm(as_completed(future_to_batch),
                                   total=len(batches),
                                   desc="Fetching sequences"):
                    try:
                        results = future.result()

                        # write results to file
                        for seqid, sequence in results:
                            out.write(f">{seqid}\n{sequence}\n")
                            sequences_written += 1

                    except Exception as e:
                        batch_num = future_to_batch[future]
                        logger.error(f"Error processing batch {batch_num}: {e}")
                        raise

            logger.info(f"Retrieved {sequences_written} sequences to {out_fasta}")

        else:
            file_type = determine_file_type(query_path)
            logger.info(f"Determined input file type: {file_type}")

            if file_type == "fasta":
                logger.info(f"Processing FASTA format from {query_path}")
                for header, sequence in write_fa_matches(query_path, matching_ids):
                    out.write(f"{header}\n{sequence}\n")
                    sequences_written += 1

            elif file_type == "fastq":
                logger.info(f"Processing FASTQ format from {query_path}")
                for header, sequence in write_fq_matches(query_path, matching_ids):
                    out.write(f"{header}\n{sequence}\n")
                    sequences_written += 1

            else:
                raise ValueError(f"Unsupported file type: {file_type}")

            logger.info(f"Successfully wrote {sequences_written} matching sequences to {out_fasta}")

    # ===============================
    # Generate sequence statistics
    # ===============================
    cmd = ["seqkit", "stats",
           out_fasta,
           "-o", stats_path,
           "-a"]

    logger.debug(f"Running command: {' '.join(cmd)}")
    subprocess.run(cmd, check=True)

    return out_fasta, stats_path





# ===============================
# aastk max_score CLI FUNCTION
# ===============================
def max_score(extracted: str,
              output_dir: str,
              matrix: str = 'BLOSUM45',
              force: bool = False):
    """
    Calculates max scores for sequences using a BLOSUM matrix.

    Args:
        extracted (str): Path to the extracted FASTA file.
        matrix (str): BLOSUM matrix name (default: 'BLOSUM45').
        output_dir (str): Directory where the output file should be stored.
        force (bool): If true, existing files/directories in output path are overwritten

    Returns:
        max_scores: Dictionary of protein headers and their max scores.
    """
    protein_name = determine_dataset_name(extracted, '.', 0, '_matched')

    # ===============================
    # Output file setup
    # ===============================
    out_file = ensure_path(output_dir, f"{protein_name}_max_scores.tsv", force=force)

    # validate requested matrix
    if matrix not in BLOSUM_DIAGONALS.keys():
        raise ValueError(f"Invalid matrix: {matrix}. Must be one of {BLOSUM_DIAGONALS.keys()}")

    logger.info(f"Calculating max scores using {matrix} matrix")

    # ===============================
    # Read input sequences
    # ===============================
    sequences = read_fasta_to_dict(extracted)
    logger.info(f"Read {len(sequences)} sequences from {extracted}")

    # ===============================
    # Calculate max scores
    # ===============================
    max_scores = {}
    for header, sequence in sequences.items():
        score = 0
        for amino_acid in sequence:
            aa = amino_acid.upper()  # just to make sure that we have consistency
            if aa in BLOSUM_DIAGONALS[matrix]:
                score += BLOSUM_DIAGONALS[matrix][amino_acid]
            else:
                logger.warning(f"Unknown amino acid '{amino_acid}' in sequence {header}")
        max_scores[header] = score

    # ===============================
    # Write results to output file
    # ===============================
    with open(out_file, 'w') as out:
        out.write("Protein_id\tmax_score\n")
        for header, score in max_scores.items():
            out.write(f"{header}\t{score}\n")

    logger.info(f"Successfully wrote max scores to {out_file}")

    return out_file





# ===============================
# aastk bsr CLI FUNCTION
# ===============================
def bsr(blast_tab: str,
        max_scores_path: str,
        output_dir: str,
        key_column: int = 0,
        column_info_path: str = None,
        score_column: int = None,
        force: bool = False):
    """
    Computes BSR (Blast Score Ratio) using a BLAST tab file and max scores from a TSV.

    Args:
        blast_tab (str): Path to DIAMOND/BLAST output file (must include 'score' column).
        max_scores_path (str): Path to TSV file with max scores (headers: Protein_id, max_score).
        output_dir (str): Directory to save the BSR results.
        key_column (int): Column index in blast_tab to use for matching. (Default: 0 for 'qseqid')
        column_info_path (str): Path to file containing column names for the blast tabular output file to retrieve index of the score column
        score_column (int): Index of blast tabular output file's score column
        force (bool): If true, existing files/directories in output path are overwritten

    Returns:
        bsr_output (str): Path to the output file with BSR values.
    """
    protein_name = determine_dataset_name(blast_tab, '.', 0, '_hits')

    # ===============================
    # Output file setup
    # ===============================
    bsr_file = ensure_path(output_dir, f"{protein_name}_bsr.tsv", force=force)

    logger.info(f"Computing blast score ratio (BSR) for {protein_name}")
    logger.info(f"Using blast tab file: {blast_tab}")
    logger.info(f"Using max scores file: {max_scores_path}")

    # ===============================================================================================
    # Determination of BLAST output column indices to be used for retrieval of relevant parameters
    # ===============================================================================================
    # retrieve indices for essential columns from column info file if path is provided
    if column_info_path and Path(column_info_path).exists():
        try:
            with open(column_info_path, 'r') as f:
                column_info = json.load(f)
                if 'score' in column_info:
                    raw_score_column = column_info['score']
                else:
                    raise ValueError("Column 'score' not found in column info file")
                if 'qlen' in column_info:
                    qlen_column = column_info['qlen']
                else:
                    raise ValueError("Column 'qlen' not found in column info file")
                if 'pident' in column_info:
                    pident_column = column_info['pident']
                else:
                    raise ValueError("Column 'pident' not found in column info file")
        except (json.JSONDecodeError, IOError) as e:
            logger.warning(f"Error reading column info file: {e}")
    # if no column info file is provided we use the input score column index
    elif score_column:
        raw_score_column = score_column - 1
        logger.info(f"Using score column index {raw_score_column} from column info file")
    else:
        logger.warning(f"Column info file not provided or not found. "
                       f"Please enter path to column info file or pass "
                       f"the index of the score column in your BLAST tabular output")

    # ===============================
    # Load max scores data
    # ===============================
    max_scores = {}
    with open(max_scores_path) as tsv:
        header = tsv.readline()
        for line in tsv:
            parts = line.strip().split('\t')
            if len(parts) >= 2:
                max_scores[parts[0]] = float(parts[1])

    logger.info(f"Loaded {len(max_scores)} max scores")

    # ===============================
    # Process BLAST results and caclulate BSR
    # ===============================
    processed_count = 0
    error_count = 0

    # write the blast score ratio output file
    with open(blast_tab) as infile, open(bsr_file, 'w') as out:
        out.write("qseqid\tsseqid\tpident\tqlen\tscore\tmax_score\tBSR\n")

        for line_num, line in enumerate(infile, 1):
            parts = line.strip().split('\t')

            if len(parts) <= max(key_column, raw_score_column):
                logger.warning(f"Line {line_num}: Not enough columns "
                               f"(expected at least {max(key_column, score_column)+1}, got {len(parts)})")
                error_count += 1
                continue

            key = parts[key_column]
            try:
                # raw score is the alignment score
                raw_score = float(parts[raw_score_column])
                # % sequence identity (used to color the datapoints in the plot later)
                pident = float(parts[pident_column])
                # query length as useful additional information
                qlen = int(parts[qlen_column])

                if key not in max_scores:
                    logger.warning(f"Line {line_num}: No max score found for '{key}'")
                    error_count += 1

                max_score = max_scores[key]
                if max_score <= 0:
                    logger.warning(f"Line {line_num}: Max score is zero or negative for '{key}': {max_score}")
                    error_count += 1
                    continue

                # calculate the blast score ratio
                bsr = raw_score / max_score
                out.write(f"{parts[0]}\t{parts[1]}\t{pident:.2f}\t{qlen}\t{raw_score:.1f}\t{max_score:.1f}\t{bsr:.4f}\n")
                processed_count += 1

            except (KeyError, ValueError, IndexError) as e:
                logger.warning(f"Line {line_num}: Error processing '{line.strip()}': {e}")
                if isinstance(e, IndexError):
                    logger.warning(
                        f"Line {line_num}: Tried to access column {score_column} but line only has {len(parts)} columns")
                error_count += 1

    logger.info(f"Successfully processed {processed_count} entries and wrote BSR to {bsr_file}")
    if error_count > 0:
        logger.warning(f"Encountered {error_count} errors during BSR calculation")
    return bsr_file





# ===============================
# aastk pasr_plot CLI FUNCTION
# ===============================
def pasr_plot(bsr_file: str,
             output_dir: str,
             yaml_path: str,
             svg: bool = False,
             force: bool = False,
             update: bool = False):
    """
    Creates a scatter plot of the BSR data flanked by histograms showing the distribution of datapoints alongside the axes.
    """
    logger = logging.getLogger(__name__)

    protein_name = determine_dataset_name(bsr_file, '.', 0, '_bsr')

    # ===============================
    # Output file path setup
    # ===============================
    if update:
        if svg:
            out_graph = ensure_path(output_dir, f'{protein_name}_updated_bsr.svg', force=force)
        else:
            out_graph = ensure_path(output_dir, f'{protein_name}_updated_bsr.png', force=force)
    else:
        if svg:
            out_graph = ensure_path(output_dir, f'{protein_name}_bsr.svg', force=force)
        else:
            out_graph = ensure_path(output_dir, f'{protein_name}_bsr.png', force=force)

    logger.info(f"Creating BSR scatter plot for {protein_name}")

    try:
        # ===============================
        # Load data
        # ===============================
        bsr_df = pd.read_csv(bsr_file, sep='\t', header=0)

        required_cols = ['max_score', 'score', 'pident', 'qlen', 'BSR']
        for col in required_cols:
            if col not in bsr_df.columns:
                raise ValueError(f"Required column '{col}' not found in BSR data")

        # ===============================
        # Define axis ranges and bins
        # ===============================
        bin_width = 10
        score_max = bsr_df['score'].max()

        # define common axis limits for scatter + histograms
        xlim = (0, 1.5 * score_max)
        ylim = (0, score_max)

        x_bins = np.arange(xlim[0], xlim[1] + bin_width, bin_width)
        y_bins = np.arange(ylim[0], ylim[1] + bin_width, bin_width)

        # ===============================
        # Layout
        # ===============================
        fig, axs = plt.subplot_mosaic(
            [['histx', '.'],
             ['scatter', 'histy']],
            figsize=(8, 8),
            width_ratios=(4, 1),
            height_ratios=(1, 4),
            layout='constrained'
        )

        # ===============================
        # Scatter plot
        # ===============================
        scatter = axs['scatter'].scatter(
            bsr_df['max_score'],
            bsr_df['score'],
            c=bsr_df['pident'],
            cmap='viridis',
            edgecolor='k',
            alpha=0.5,
            vmin=0,
            vmax=100
        )
        axs['scatter'].set_xlabel('Calculated maximum score')
        axs['scatter'].set_ylabel('Alignment score to seed set')
        axs['scatter'].set_xlim(xlim)
        axs['scatter'].set_ylim(bottom=0)

        cb_ax = inset_axes(axs['scatter'], width="5%", height="30%", loc='upper left', borderpad=1)
        cbar = fig.colorbar(scatter, cax=cb_ax)
        cbar.set_label('% sequence identity')

        # ===============================
        # Histograms aligned with scatter
        # ===============================
        # Top histogram
        x_hist, x_edges = np.histogram(bsr_df['max_score'], bins=x_bins)
        axs['histx'].bar(
            x_edges[:-1],
            x_hist,
            width=bin_width,
            align='edge',
            color='black'
        )
        axs['histx'].set_xlim(xlim)
        axs['histx'].set_ylabel('Counts')
        axs['histx'].set_yticks([0, round(max(x_hist)/2), round(max(x_hist))])
        axs['histx'].tick_params(labelbottom=False)
        axs['histx'].set_title(f'Protein Alignment Score Ratio for {protein_name}')

        # Right histogram
        y_hist, y_edges = np.histogram(bsr_df['score'], bins=y_bins)
        axs['histy'].barh(
            y_edges[:-1],
            y_hist,
            height=bin_width,
            align='edge',
            color='black'
        )
        axs['histy'].set_ylim(bottom=0)
        axs['histy'].set_xlabel('Counts')
        axs['histy'].set_xticks([0, round(max(y_hist)/2), round(max(y_hist))])
        axs['histy'].tick_params(labelleft=False)


        # ===============================
        # Threshold lines (update mode)
        # ===============================
        if update:
            try:
                with open(yaml_path) as f:
                    thresholds = yaml.safe_load(f)
            except Exception as e:
                raise RuntimeError(f"Failed to load thresholds from {yaml_path}: {e}") from e

            max_score_min = thresholds.get("max_score_min", 0)
            max_score_max = thresholds.get("max_score_max", float('inf'))
            dbmin = thresholds.get("dbmin", None)
            bsr_min = thresholds.get("bsr", None)

            if max_score_min is not None:
                axs['scatter'].axvline(max_score_min, color='black', linestyle='--', linewidth=1.0)
            if max_score_max is not None:
                axs['scatter'].axvline(max_score_max, color='black', linestyle='--', linewidth=1.0)
            if dbmin is not None:
                axs['scatter'].axhline(dbmin, color='black', linestyle='--', linewidth=1.0)
            if bsr_min is not None:
                x_vals = np.linspace(xlim[0], xlim[1], 500)
                y_vals = bsr_min * x_vals
                axs['scatter'].plot(x_vals, y_vals, color='black', linestyle='--', linewidth=1.0)

        # ===============================
        # Save
        # ===============================
        fig.savefig(out_graph, dpi=300)
        plt.close(fig)

        logger.info(f"Successfully created BSR plot at {out_graph}")
        return out_graph

    except Exception as e:
        logger.error(f"Error creating BSR plot: {e}")
        raise RuntimeError(f"Failed to create BSR plot: {e}") from e





# ===============================
# aastk pasr_select CLI FUNCTION
# ===============================
def pasr_metadata(max_score_min: int,
                  max_score_max: int,
                  output_dir: str,
                  dbmin: int = None,
                  bsr: float = None,
                  seed: str = None,
                  bsr_file: str = None,
                  force: bool = False):
    """
    Writes metadata parameters to a YAML file.

    Args:
        max_score_min: Minimum max_score-score threshold.
        max_score_max: Maximum max_score-score threshold.
        output_dir: Directory to save the metadata file.
        dbmin: Minimum database score threshold.
        bsr: Minimum BSR threshold.
        seed (str): Path to seed FASTA.
        bsr_file (str): Path to BSR tab separated file.
        force (bool): If true, existing files/directories in output path are overwritten.

    Returns:
        Path to the metadata file.
    """
    logger.info("Starting metadata creation")

    # ===============================
    # Parameter validation
    # ===============================
    if seed is None and bsr_file is None:
        error_msg = "Either seed or bsr_file must be provided"
        logger.error(error_msg)
        raise ValueError(error_msg)

    if seed is not None and bsr_file is not None:
        error_msg = "seed and bsr_file are mutually exclusive. Provide only one."
        logger.error(error_msg)
        raise ValueError(error_msg)

    # ===============================
    # Protein name determination
    # ===============================
    if seed is not None:
        logger.info(f"Determining protein name from seed file: {seed}")
        seed_name = Path(seed).name
        protein_name = determine_dataset_name(seed_name, '.', 0)
        logger.info(f"Extracted protein name '{protein_name}' from seed file")
    elif bsr_file is not None:
        logger.info(f"Determining protein name from BSR file: {bsr_file}")
        protein_name = determine_dataset_name(bsr_file, '.', 0, '_bsr')
        logger.info(f"Extracted protein name '{protein_name}' from BSR file")
    else:
        error_msg = "Unexpected state: both seed and bsr_file are None"
        logger.error(error_msg)
        raise RuntimeError(error_msg)

    # ===============================
    # Output file path setup
    # ===============================
    try:
        yaml_path = ensure_path(output_dir, f"{protein_name}.yaml", force=force)
        logger.info(f"Writing metadata for {protein_name} to {yaml_path}")
    except Exception as e:
        logger.error(f"Failed to create output path: {e}")
        raise

    # ===============================
    # Parameter collection and validation
    # ===============================
    logger.debug(f"Collecting parameters: max_score_min={max_score_min}, max_score_max={max_score_max}, "
                 f"dbmin={dbmin}, bsr={bsr}")

    # Validate threshold values
    if max_score_min < 0 or max_score_max < 0:
        error_msg = "max_score_min and max_score_max must be non-negative"
        logger.error(error_msg)
        raise ValueError(error_msg)

    if max_score_min > max_score_max:
        error_msg = f"max_score_min ({max_score_min}) cannot be greater than max_score_max ({max_score_max})"
        logger.error(error_msg)
        raise ValueError(error_msg)

    if bsr is not None and (bsr < 0 or bsr > 1):
        error_msg = f"BSR value must be between 0 and 1, got {bsr}"
        logger.error(error_msg)
        raise ValueError(error_msg)

    if dbmin is not None and dbmin < 0:
        error_msg = f"dbmin must be non-negative, got {dbmin}"
        logger.error(error_msg)
        raise ValueError(error_msg)

    params = {
        "protein_name": protein_name,
        "max_score_min": max_score_min,
        "max_score_max": max_score_max,
        "dbmin": dbmin,
        "bsr": bsr
    }

    # remove None values
    params = {k: v for k, v in params.items() if v is not None}
    logger.debug(f"Final parameters to write: {params}")

    # ===============================
    # Write YAML file
    # ===============================
    try:
        yaml_str = yaml.dump(params)
        logger.debug(f"Generated YAML content:\n{yaml_str}")

        with open(yaml_path, "w") as f:
            f.write(yaml_str)

        logger.info(f"Successfully wrote metadata to {yaml_path}")
        return yaml_path

    except Exception as e:
        logger.error(f"Failed to write metadata file {yaml_path}: {e}")
        raise RuntimeError(f"Failed to write metadata file: {e}") from e
    
def pasr_select(yaml_path: str,
                matched_fasta: str,
                bsr_table: str,
                output_dir: str,
                max_score_min: int = None,
                max_score_max: int = None,
                dbmin: int = None,
                bsr: float = None,
                force: bool = False,
                create_yaml: bool = False,
                params: bool = False):
    """
    Subsets matched sequences based on YAML thresholds or provided parameters.

    Args:
        yaml_path(str): Path to the metadata yaml file.
        matched_fasta (str): Path to the matched sequences FASTA.
        bsr_table (str): Path to the BSR table.
        output_dir (str): Directory to save the subsetted sequences.
        max_score_min (int): Minimum max_score score threshold (required when params=True).
        max_score_max (int): Maximum max_score score threshold (required when params=True).
        dbmin (int): Minimum database score threshold (mutually exclusive with bsr).
        bsr (float): Minimum BSR threshold (mutually exclusive with dbmin).
        force (bool): If true, existing files/directories in output path are overwritten.
        create_yaml (bool): Create a YAML file with the provided parameters.
        params (bool): Use provided parameters instead of YAML file (mutually exclusive with yaml_path).
    """
    # check if seqkit is in path
    check_dependency_availability('seqkit')

    logger.info("Starting sequence subsetting based on thresholds")

    protein_name = determine_dataset_name(bsr_table, '.', 0, '_bsr')

    # ===============================
    # Parameter validation
    # ===============================
    # check mutual exclusivity of yaml_path and params
    if yaml_path is not None and params:
        raise ValueError(
            "yaml_path and params are mutually exclusive. Use either a YAML file or provide parameters directly.")

    if yaml_path is None and not params:
        raise ValueError("Either yaml_path must be provided or params must be True.")

    created_yaml_path = None

    # ===============================
    # Threshold loading/validation
    # ===============================
    if params:
        # validate required parameters when using params=True
        if max_score_min is None or max_score_max is None:
            raise ValueError("max_score_min and max_score_max are required when params=True")

        # ensure either dbmin or bsr is provided (but not both)
        if dbmin is None and bsr is None:
            raise ValueError("Either dbmin or bsr must be provided when params=True")

        if dbmin is not None and bsr is not None:
            raise ValueError("dbmin and bsr are mutually exclusive. Provide only one.")

        logger.info("Using provided parameters for thresholds")

        # set the threshold values
        bsr_min = bsr

        # create YAML file if requested
        if create_yaml:
            logger.info("Creating YAML file with provided parameters")
            created_yaml_path = pasr_metadata(
                max_score_min=max_score_min,
                max_score_max=max_score_max,
                output_dir=output_dir,
                dbmin=dbmin,
                bsr=bsr,
                bsr_file=bsr_table,
                force=force
            )

    else:
        # load thresholds from YAML file
        logger.info(f"Loading thresholds from YAML file: {yaml_path}")
        try:
            with open(yaml_path) as f:
                thresholds = yaml.safe_load(f)
        except Exception as e:
            logger.error(f"Failed to load YAML file {yaml_path}: {e}")
            raise RuntimeError(f"Failed to load thresholds from {yaml_path}: {e}") from e

        # get thresholds and assign default values
        protein_name = thresholds.get("protein_name", None)
        max_score_min = thresholds.get("max_score_min", 0)
        max_score_max = thresholds.get("max_score_max", float('inf'))
        dbmin = thresholds.get("dbmin", None)
        bsr_min = thresholds.get("bsr", None)

        if protein_name is None:
            raise ValueError("protein_name must be specified in the YAML file")

    # ===============================
    # Output file path setup
    # ===============================
    # create output file path
    output_file = f"{protein_name}_matched_update.faa"
    stats_file = f"{protein_name}_matched_update.stats"
    output_path = ensure_path(output_dir, output_file, force=force)
    stats_path = ensure_path(output_dir, stats_file, force=force)

    # ===============================
    # Data loading and filtering
    # ===============================
    # load BSR table
    bsr_df = pd.read_csv(bsr_table, sep='\t')

    # apply filters defined by the required arguments max_score_min and max_score_max to the BSR table
    filtered = bsr_df[(bsr_df['max_score'] >= max_score_min) & (bsr_df['max_score'] <= max_score_max)]

    # apply mutually exclusive filters to the BSR table (either min. alignment score or min. BSR)
    if dbmin is not None:
        filtered = filtered[filtered['score'] >= dbmin]
        logger.info(f"Applied dbmin filter: score >= {dbmin}")
    elif bsr_min is not None:
        filtered = filtered[filtered['BSR'] >= bsr_min]
        logger.info(f"Applied BSR filter: BSR >= {bsr_min}")

    if filtered.empty:
        logger.warning("No sequences passed the filtering criteria!")

    # extract sequence IDs kept after filtering
    filtered_ids = set(filtered['qseqid'])

    # ===============================
    # Sequence filtering and writing
    # ===============================
    # load matched FASTA
    sequences = read_fasta_to_dict(matched_fasta)

    # track how many sequences we are keeping
    kept = 0

    # retrieve filtered sequences from original homologous faa file
    with open(output_path, 'w') as out:
        for header, seq in sequences.items():
            if header in filtered_ids:
                out.write(f">{header}\n{seq}\n")
                kept += 1

    logger.info(f"Filtered sequences written to {output_path}: "
                f"{kept} sequences out of {len(sequences)} original sequences")

    # ===============================
    # Statistics generation
    # ===============================
    # generate seqkit stats for the filtered fasta file
    cmd = ["seqkit", "stats",
           output_path,
           "-o", stats_path,
           "-a"]

    logger.debug(f"Running command: {' '.join(cmd)}")
    subprocess.run(cmd, check=True)

    # ===============================
    # Generate updated BSR plot
    # ===============================
    if created_yaml_path:
        yaml_for_plot = created_yaml_path
    elif yaml_path:
        yaml_for_plot = yaml_path
    else:
        yaml_for_plot = None

    if yaml_for_plot:
        pasr_plot(
            bsr_file=bsr_table,
            output_dir=output_dir,
            yaml_path=yaml_for_plot,
            force=force,
            update=True
        )

    if created_yaml_path:
        return output_path, stats_path, created_yaml_path
    else:
        return output_path, stats_path





# ===============================
# aastk pasr WORKFLOW
# ===============================
def pasr(seed_fasta: str,
         query_fasta: str,
         output_dir: str,
         sensitivity: str,
         db_path: str,
         block: int,
         chunk: int,
         matrix: str = 'BLOSUM45',
         key_column: int = 0,
         threads: int = 1,
         update: bool = False,
         yaml_path: str = None,
         sql: bool = False,
         keep: bool = False,
         svg: bool = False,
         force: bool = False):
    """
    PASR workflow with configurable output directory.
    Runs:
    aastk build
    aastk search
    aastk get_hit_seqs
    aastk max_score
    aastk bsr
    aastk pasr_plot

    Optional (with provided metadata YAML file):
    aastk pasr_select
    aastk pasr_plot

    Args:
        seed_fasta (str): Path to seed FASTA.
        query_fasta (str): Path to query FASTA.
        matrix (str): BLOSUM matrix ('BLOSUM45' or 'BLOSUM62').
        output_dir (str): Output directory (default: current directory).
        sensitivity (str): Choose sensitivity of diamond blastp search (Default: --fast)
        block (int): Choose diamond blastp sequence block size in billions of letters. (Default: 6)
        chunk (int): Choose number of chunks for diamond blastp index processing. (Default: 2)
        key_column: Column index in the BLAST tab file to pull unique IDs from (default is 0).
        threads (int): Number of threads (default: 1).
        update (bool): Creates updated matched_fasta and bsr plot using metadata yaml file
        yaml_path (str): Path to metadata yaml file
        svg (bool): If true, plots will be generated in SVG format
        force (bool): If true, existing files/directories in output path are overwritten
    """
    # ===============================
    # Parameter validation and setup
    # ===============================
    if update and not yaml_path:
        logger.error("YAML path is required if update is True")
        exit()

    if matrix not in BLOSUM_DIAGONALS.keys():
        raise ValueError(f"Invalid matrix: {matrix}. Must be one of {BLOSUM_DIAGONALS.keys()}")

    seed_name = Path(seed_fasta).name
    protein_name = determine_dataset_name(seed_name, '.', 0)

    # check for output_dir
    logger.info(f"Running PASR workflow for {protein_name}")
    logger.info(f"Output directory: {output_dir}")

    # store all the output paths in dictionaries
    intermediate_results = {}
    results = {}

    # ===============================
    # Database building
    # ===============================
    try:
        logger.info("Building protein database")
        diamond_db_path = build(seed_fasta, threads, output_dir, force=force)
        intermediate_results['db_path'] = f"{db_path}.dmnd"

        # ===============================
        # Database search
        # ===============================
        logger.info("Searching protein database")
        search_output, column_info_path = search(diamond_db_path, query_fasta, threads, output_dir, sensitivity, block, chunk, force=force)
        intermediate_results['search_output'] = search_output
        intermediate_results['column_info_path'] = column_info_path

        # ===============================
        # Sequence extraction
        # ===============================
        logger.info("Extracting matching sequences")
        matched_fasta, stats_path = get_hit_seqs(search_output, query_fasta, output_dir, db_path, key_column, threads=threads, sql=sql, force=force)
        results['matched_fasta'] = matched_fasta
        results['stats_path'] = stats_path

        # ===============================
        # Score calculations
        # ===============================
        logger.info("Calculating max scores")
        max_scores = max_score(matched_fasta, output_dir, matrix, force=force)
        intermediate_results['max_scores'] = max_scores

        logger.info("Calculating blast score ratios")
        bsr_file = bsr(search_output, max_scores, output_dir, key_column, column_info_path, score_column=None, force=force)
        results['bsr_file'] = bsr_file

        # ===============================
        # Visualization
        # ===============================
        logger.info("Creating BSR plot")
        bsr_plot = pasr_plot(bsr_file, output_dir, yaml_path, svg=svg, force=force, update=False)
        results['bsr_plot'] = bsr_plot

        # ===============================
        # Update workflow (optional)
        # ===============================
        if update:
            logger.info("Running update for specified data")
            subset_fasta, update_stats_path = pasr_select(yaml_path, matched_fasta, bsr_file, output_dir, force=force)
            results['subset_fasta'] = subset_fasta
            results['update_stats_path'] = update_stats_path
            updated_plot = pasr_plot(bsr_file, output_dir,  yaml_path, svg=svg, force=force, update=update)
            results['updated_plot'] = updated_plot

        logger.info("PASR workflow completed successfully")
        return results

    except Exception as e:
        logger.error(f"PASR workflow failed: {e}")
        exit()
    finally:
        if not keep and intermediate_results:
            logger.info("Cleaning up intermediate files")
            for key, filepath in intermediate_results.items():
                try:
                    file_path = Path(filepath)
                    if file_path.exists():
                        file_path.unlink()
                        logger.debug(f"Deleted {filepath}")
                except Exception as e:
                    logger.warning(f"Failed to delete {filepath}: {e}")



from .cugo import *
from .util import *

import sqlite3
import subprocess
from tqdm import tqdm
from concurrent.futures import ThreadPoolExecutor, as_completed
from textwrap import dedent
import logging
import gzip
from pathlib import Path
import shutil


logger = logging.getLogger(__name__)

# define database table schema
PROTEIN_SCHEMA = {
    'seqID': 'TEXT PRIMARY KEY',
    'parent_ID': 'TEXT',
    'aa_length': 'INTEGER',
    'strand': 'TEXT',
    'COG_ID': 'TEXT',
    'KEGG_ID': 'TEXT',
    'Pfam_ID': 'TEXT',
    'cugo_number': 'INTEGER',
    'no_tmh': 'INTEGER',
    'protein_seq': 'BLOB'
}

GENOME_SCHEMA = {
    'genome_ID': 'TEXT PRIMARY KEY',
    'domain': 'TEXT',
    'phylum': 'TEXT',
    'class': 'TEXT',
    'order_tax': 'TEXT',
    'family': 'TEXT',
    'genus': 'TEXT',
    'species': 'TEXT',
    'culture_collection': 'INT',
    'animal_associated': 'REAL',
    'aquatic': 'REAL',
    'built': 'REAL',
    'other': 'REAL',
    'sediment': 'REAL',
    'soil': 'REAL',
    'unassigned_high_level': 'REAL',
    'human': 'REAL',
    'invertebrate': 'REAL',
    'other_vertebrate': 'REAL',
    'unspecified_animal': 'REAL',
    'aquatic_other': 'REAL',
    'aquatic_unspecified': 'REAL',
    'freshwater': 'REAL',
    'groundwater': 'REAL',
    'marine': 'REAL',
    'built_other': 'REAL',
    'drinking_water': 'REAL',
    'wastewater': 'REAL',
    'air': 'REAL',
    'bacteria': 'REAL',
    'eukaryote_other': 'REAL',
    'food': 'REAL',
    'geothermal': 'REAL',
    'hypersaline': 'REAL',
    'other_unspecified': 'REAL',
    'plant_associated': 'REAL',
    'subsurface': 'REAL',
    'synthetic': 'REAL',
    'viral': 'REAL',
    'freshwater_sediment': 'REAL',
    'marine_sediment': 'REAL',
    'sediment_unspecified': 'REAL',
    'desert': 'REAL',
    'forest': 'REAL',
    'rhizosphere': 'REAL',
    'soil_agricultural': 'REAL',
    'soil_other': 'REAL',
    'soil_unspecified': 'REAL',
    'tundra_wetland': 'REAL',
    'unassigned_low_level': 'REAL'
}

# extract column name lists for queries (same order)
PROTEIN_COLUMNS = list(PROTEIN_SCHEMA.keys())
GENOME_COLUMNS = list(GENOME_SCHEMA.keys())

# categorize columns for later use
BASE_COLUMNS = ['seqID', 'parent_ID', 'aa_length', 'strand']

ANNOTATION_COLUMNS = ['COG_ID', 'KEGG_ID', 'Pfam_ID']

TAXONOMY_COLUMNS = [
    'genome_ID', 'domain', 'phylum', 'class',
    'order_tax', 'family', 'genus', 'species'
]

CULTURE_COLLECTION_COLUMNS = ['culture_collection']

HIGH_LEVEL_ENV_COLUMNS = [
    'animal_associated', 'aquatic', 'built', 'other',
    'sediment', 'soil', 'unassigned_high_level'
]

LOW_LEVEL_ENV_COLUMNS = [
    'human', 'invertebrate', 'other_vertebrate', 'unspecified_animal',
    'aquatic_other', 'aquatic_unspecified', 'freshwater', 'groundwater', 'marine',
    'built_other', 'drinking_water', 'wastewater',
    'air', 'bacteria', 'eukaryote_other', 'food', 'geothermal', 'hypersaline',
    'other_unspecified', 'plant_associated', 'subsurface', 'synthetic', 'viral',
    'freshwater_sediment', 'marine_sediment', 'sediment_unspecified',
    'desert', 'forest', 'rhizosphere', 'soil_agricultural', 'soil_other',
    'soil_unspecified', 'tundra_wetland', 'unassigned_low_level'
]

# ====================================
#
#   SQLite database setup
#
# ====================================
def setup_database(db_path: str) -> sqlite3.Connection:
    conn = sqlite3.connect(db_path)

    protein_cols = ',\n                     '.join(
        [f'{col} {dtype}' for col, dtype in PROTEIN_SCHEMA.items()]
    )

    conn.execute(f'''
                 CREATE TABLE IF NOT EXISTS protein_data (
                    {protein_cols}
                 )
                 ''')

    genome_cols = ',\n                     '.join(
        [f'{col} {dtype}' for col, dtype in GENOME_SCHEMA.items()]
    )

    conn.execute(f'''
                 CREATE TABLE IF NOT EXISTS genome_data (
                    {genome_cols}  
                 )
                 ''')

    view_cols = [col for col in PROTEIN_COLUMNS if col != 'protein_seq']
    view_select = ',\n                     '.join(view_cols)

    conn.execute(f'''
                 CREATE VIEW IF NOT EXISTS protein_data_readable AS
                 SELECT
                     {view_select},
                     CASE WHEN protein_seq IS NULL THEN NULL ELSE '<COMPRESSED_BLOB>' END as protein_seq_status
                 FROM protein_data
                 ''')


    conn.commit()
    return conn

# ============================================================ #
# GFF parser                                                   #
# ============================================================ #
def extract_gene_info(line: list) -> tuple:
    """
    Extracts gene information from a GFF/GTF annotation line.

    Args:
        line (list): parsed annotation line with genomic feature data

    Returns:
        seqID (str): sequence identifier with underscores normalized
        annotation_ID (str): annotation identifier or 'NA' if not found
        parent (str): chromosome/contig name
        direction (str): strand direction (+/-)
        gene_start (str): start position
        gene_end (str): end position
        nuc_length (int): nucleotide sequence length
        aa_length (int): amino acid sequence length
    """
    # extract basic genomic coordinates and features
    parent = line[0]  # chromosome/contig name
    direction = line[6]  # strand direction (+/-)
    gene_start = line[3]  # start position
    gene_end = line[4]  # end position
    attributes = line[8]

    # parse attributes field
    attr_dict = {}
    for attr in attributes.split(';'):
        if '=' in attr:
            key, value = attr.split('=', 1)
            attr_dict[key] = value

    seqID = attr_dict.get('ID', '')
    annotation_ID = attr_dict.get('Name', '')

    if not annotation_ID:
        annotation_ID = 'NA'

    # calculate sequence lengths
    nuc_length = abs(int(gene_end) - int(gene_start)) + 1  # nucleotide length
    aa_length = int(nuc_length / 3)  # amino acid length
    return seqID, annotation_ID, parent, direction, gene_start, gene_end, nuc_length, aa_length


def cugo_boundaries(direction: str,
                    prev_direction: str,
                    next_direction: str,
                    parent: str,
                    prev_parent: str,
                    next_parent: str,
                    cugo_count: int,
                    cugo_size: dict,
                    cugo_size_count: int,
                    prev_cugo: int = None):
    """
    Determines CUGO (Co-oriented Unidirectional Gene Order) boundaries and classifications.

    Args:
        direction (str): current gene strand direction (+/-)
        prev_direction (str): previous gene strand direction (+/-)
        next_direction (str): next gene strand direction (+/-)
        parent (str): current gene's chromosome/contig name
        prev_parent (str): previous gene's chromosome/contig name
        next_parent (str): next gene's chromosome/contig name
        cugo_count (int): running count of CUGO regions
        cugo_size (dict): mapping of CUGO IDs to their sizes
        cugo_size_count (int): running count of genes in current CUGO
        prev_cugo (int, optional): previous CUGO identifier

    Returns:
        cugo (int): current CUGO identifier
        cugo_start (str): start boundary type ('sequence_edge', 'strand_change', or 'NA')
        cugo_end (str): end boundary type ('sequence_edge', 'strand_change', or 'NA')
        cugo_count (int): updated CUGO count
        cugo_size_count (int): updated gene count for current CUGO
    """
    cugo, cugo_start, cugo_end = 'NA', 'NA', 'NA'

    # middle of cugo - same direction and parent for all three genes
    if (direction == prev_direction == next_direction) and (parent == prev_parent == next_parent):
        cugo = prev_cugo
        cugo_size_count += 1

    # start of contig/scaffold - different parent from previous gene
    elif parent != prev_parent:
        cugo = cugo_count
        cugo_size_count = 1

        if direction == '+':
            cugo_start = 'sequence_edge'
            if parent != next_parent:
                # single gene contig
                cugo_end = 'sequence_edge'
                cugo_size[cugo] = cugo_size_count
                cugo_size_count = 0
                cugo_count += 1
            elif direction != next_direction:
                # strand change at start of contig
                cugo_end = 'strand_change'
                cugo_size[cugo] = cugo_size_count
                cugo_size_count = 0
                cugo_count += 1
        else:  # direction == '-'
            cugo_end = 'sequence_edge'
            if parent != next_parent:
                # single gene contig
                cugo_start = 'sequence_edge'
                cugo_size[cugo] = cugo_size_count
                cugo_size_count = 0
                cugo_count += 1
            elif direction != next_direction:
                # strand change at start of contig
                cugo_start = 'strand_change'
                cugo_size[cugo] = cugo_size_count
                cugo_size_count = 0
                cugo_count += 1

    # end of contig/scaffold - different parent from next gene
    elif parent != next_parent:
        if direction == prev_direction:
            # continuing previous cugo
            cugo = prev_cugo
            cugo_size_count += 1
            if direction == '+':
                cugo_end = 'sequence_edge'
            else:
                cugo_start = 'sequence_edge'
            cugo_size[cugo] = cugo_size_count
            cugo_size_count = 0
            cugo_count += 1
        else:
            # new cugo at end of contig
            cugo = cugo_count
            cugo_size_count = 1
            if direction == '+':
                cugo_start = 'strand_change'
                cugo_end = 'sequence_edge'
            else:
                cugo_start = 'sequence_edge'
                cugo_end = 'strand_change'
            cugo_size[cugo] = cugo_size_count
            cugo_size_count = 0
            cugo_count += 1

    # strand change - different direction from previous or next gene
    elif direction != prev_direction or direction != next_direction:
        if direction != prev_direction and direction == next_direction:
            # start of new cugo due to strand change
            cugo = cugo_count
            cugo_size_count = 1
            if direction == '+':
                cugo_start = 'strand_change'
            else:
                cugo_end = 'strand_change'
        elif direction == prev_direction and direction != next_direction:
            # end of current cugo due to strand change
            cugo = prev_cugo
            cugo_size_count += 1
            if direction == '+':
                cugo_end = 'strand_change'
            else:
                cugo_start = 'strand_change'
            cugo_size[cugo] = cugo_size_count
            cugo_size_count = 0
            cugo_count += 1
        else:  # direction != prev_direction and direction != next_direction
            # isolated gene with different direction from both neighbors
            cugo = cugo_count
            cugo_size_count = 1
            cugo_start = cugo_end = 'strand_change'
            cugo_size[cugo] = cugo_size_count
            cugo_size_count = 0
            cugo_count += 1

    return cugo, cugo_start, cugo_end, cugo_count, cugo_size_count

def parse_single_gff(gff_content: str) -> list:
    """
    Parses GFF content and extracts gene information with CUGO boundary analysis.

    Args:
        gff_content (str): GFF format file content as string

    Returns:
        list: List of processed gene records, each containing:
            - seqID (str): gene/sequence identifier
            - parent_ID (str): chromosome/contig identifier
            - aa_length (int): amino acid sequence length
            - strand (str): gene strand direction (+/-)
            - annotation ID (str): functional category identifier
            - cugo_number (int): CUGO region identifier
    """
    reformat_data = []
    cugo_size = {}
    cugo_count = prev_direction = prev_parent = prev_feat_type = prev_cugo = 0
    cugo_size_count = 0
    prev_line = None
    cds_count = 0

    # split content into individual lines for processing
    lines = gff_content.strip().split('\n')

    for line_count, line in enumerate(lines, 1):
        # parse tab-delimited gff line
        clean_line = line.strip().split('\t')

        # skip malformed lines (gff requires 9 columns)
        if len(clean_line) != 9:
            continue

        # extract feature type and count cds features
        feat_type = clean_line[2]
        if feat_type == 'CDS':
            cds_count += 1

        # only process cds features or transitions from cds
        if feat_type != 'CDS' and prev_feat_type != 'CDS':
            continue
        prev_feat_type = feat_type

        # initialize with first cds line
        if prev_line is None:
            prev_line = clean_line
            continue

        # extract gene information from previous line
        seqID, annotation_ID, parent, direction, gene_start, gene_end, nuc_length, aa_length = extract_gene_info(
            prev_line)

        # get next gene's parent and direction for cugo boundary analysis
        next_parent = clean_line[0]
        next_direction = clean_line[6]

        # determine cugo boundaries and classifications
        cugo, cugo_start, cugo_end, cugo_count, cugo_size_count = cugo_boundaries(
            direction, prev_direction, next_direction,
            parent, prev_parent, next_parent,
            cugo_count, cugo_size, cugo_size_count, prev_cugo
        )

        # store processed gene data (all 11 fields for database)
        reformat_data.append([seqID, parent, aa_length,
                              direction, annotation_ID, cugo])

        # update tracking variables for next iteration
        prev_line = clean_line
        prev_direction = direction
        prev_parent = parent
        prev_cugo = cugo

    # process the last gene in the file (no next gene to compare)
    if len(clean_line) == 9 and clean_line[2] == 'CDS':
        seqID, annotation_ID, parent, direction, gene_start, gene_end, nuc_length, aa_length = extract_gene_info(
            clean_line)

        # handle last gene - no next gene, so use none for next_parent
        cugo, cugo_start, cugo_end, cugo_count, cugo_size_count = cugo_boundaries(
            direction, prev_direction, None,  # no next direction for last gene
            parent, prev_parent, None,  # no next parent for last gene
            cugo_count, cugo_size, cugo_size_count, prev_cugo
        )

        # store final gene data (all 6 fields for database)
        reformat_data.append([seqID, parent, aa_length,
                              direction, annotation_ID, cugo])

        # finalize cugo size tracking
        cugo_size[cugo] = cugo_size_count

    # Return complete data for database storage
    return reformat_data


def process_gff_file(filepath: str):
    """
    Process a single GFF file (gzipped or plain) from disk and parse it.

    Args:
        filepath (str): Full path to the GFF file.

    Returns:
        list: Parsed gene records.
    """
    try:
        filepath = Path(filepath)

        # Open gzipped or plain text
        if filepath.suffix == ".gz":
            with gzip.open(filepath, "rt") as f:
                content = f.read()
        else:
            with open(filepath, "r", encoding="utf-8") as f:
                content = f.read()

        return parse_single_gff(content)

    except Exception as e:
        logger.warning(f"Failed to process {filepath}: {e}")
        return []


def process_tmhmm_file(tmhmm_filepath):
    try:
        with open(tmhmm_filepath, 'r') as f:
            tmhmm_data = []
            for line in f:
                if line.startswith('prot_ID'):
                    continue
                parts = line.strip().split('\t')
                if len(parts) >= 2:
                    prot_id = parts[0]
                    try:
                        no_tmh = int(parts[1])
                    except ValueError:
                        no_tmh = 0
                    tmhmm_data.append((no_tmh, prot_id))
        return tmhmm_data
    except Exception:
        return []

def populate_taxonomy_table(conn: sqlite3.Connection,
                            taxonomy_filepath: str):
    try:
        filepath = Path(taxonomy_filepath)

        if filepath.suffix == '.gz':
            file_handle = gzip.open(filepath, 'rt')
        else:
            file_handle = open(filepath, 'r')

        with file_handle as f:
            next(f)

            taxonomy_data = []
            for line in f:
                parts = line.strip().split('\t')
                if len(parts) < len(TAXONOMY_COLUMNS):
                    continue

                row_values = tuple(parts[i] for i in range(len(TAXONOMY_COLUMNS)))

                taxonomy_data.append(row_values)

            placeholders = ', '.join(['?' for _ in TAXONOMY_COLUMNS])
            columns = ', '.join(TAXONOMY_COLUMNS)

            conn.executemany(f"""
                INSERT OR REPLACE INTO genome_data
                ({columns})
                VALUES ({placeholders}) 
            """, taxonomy_data)

        conn.commit()
        logger.info(f"Inserted taxonomy data for {len(taxonomy_data)} genomes")

    except Exception as e:
        logger.error(f"Error populating taxonomy data: {e}")

def populate_culture_collection(conn: sqlite3.Connection,
                                cc_file: str):
    try:
        with open(cc_file, 'r') as f:
            next(f)

            culture_collection_data = []
            for line in f:
                parts = line.strip().split('\t')
                if len(parts) < 2:
                    continue

                genome_id = parts[0]
                culture_collection = int(parts[1])

                culture_collection_data.append((genome_id, culture_collection))

            query = """
                INSERT INTO genome_data (genome_ID, culture_collection)
                VALUES (?, ?)
                ON CONFLICT(genome_ID) DO UPDATE SET
                    culture_collection = excluded.culture_collection
            """

            conn.executemany(query, culture_collection_data)

            conn.commit()

    except Exception as e:
        logger.error(f"Error populating culture collection data: {e}")

def populate_high_level_environment(conn: sqlite3.Connection,
                                    high_level_environment_file: str):
    try:
        with open(high_level_environment_file, 'r') as f:
            next(f)

            high_level_environment_data = []
            for line in f:
                parts = line.strip().split('\t')
                if len(parts) < len(HIGH_LEVEL_ENV_COLUMNS) + 1:
                    continue

                genome_id = parts[0]

                env_values = [genome_id]
                for i, col in enumerate(HIGH_LEVEL_ENV_COLUMNS, start=1):
                    value = float(parts[i]) if parts[i].strip() else None
                    env_values.append(value)

                high_level_environment_data.append(tuple(env_values))

            columns = ['genome_ID'] + HIGH_LEVEL_ENV_COLUMNS
            placeholders = ', '.join(['?' for _ in columns])
            column_names = ', '.join(columns)
            update_clause = ', '.join([f'{col} = excluded.{col}' for col in HIGH_LEVEL_ENV_COLUMNS])

            query = f"""
                INSERT INTO genome_data ({column_names})
                VALUES ({placeholders})
                ON CONFLICT(genome_ID) DO UPDATE SET
                    {update_clause}
            """

            conn.executemany(query, high_level_environment_data)
            conn.commit()
            logger.info(f"Processed high level environment data for {len(high_level_environment_data)} genomes")

    except Exception as e:
        logger.error(f"Error populating high level environment data: {e}")


def populate_low_level_environment(conn: sqlite3.Connection,
                                   low_level_environment_file: str):
    try:
        with open(low_level_environment_file, 'r') as f:
            next(f)

            low_level_environment_data = []
            for line in f:
                parts = line.strip().split('\t')
                if len(parts) < len(LOW_LEVEL_ENV_COLUMNS) + 1:
                    continue

                genome_id = parts[0]

                env_values = [genome_id]
                for i, col in enumerate(LOW_LEVEL_ENV_COLUMNS, start=1):
                    value = float(parts[i]) if parts[i].strip() else None
                    env_values.append(value)

                low_level_environment_data.append(tuple(env_values))

        columns = ['genome_ID'] + LOW_LEVEL_ENV_COLUMNS
        placeholders = ', '.join(['?' for _ in columns])
        column_names = ', '.join(columns)
        update_clause = ', '.join([f'{col} = excluded.{col}' for col in LOW_LEVEL_ENV_COLUMNS])

        query = f"""
            INSERT INTO genome_data ({column_names})
            VALUES ({placeholders})
            ON CONFLICT(genome_ID) DO UPDATE SET
                {update_clause}
        """

        conn.executemany(query, low_level_environment_data)
        conn.commit()
        logger.info(f"Processed low level environment data for {len(low_level_environment_data)} genomes")

    except Exception as e:
        logger.error(f"Error populating low level environment data: {e}")

def stream_all_proteins(fasta: str,
                        batch_size: int = 10000):
    try:
        filepath = Path(fasta)

        if filepath.suffix == '.gz':
            file_handle = gzip.open(filepath, 'rt')
        else:
            file_handle = open(filepath, 'r')

        batch = []
        current_id = None

        with file_handle as f:
            for line in f:
                line = line.strip()

                if not line:
                    continue

                if line.startswith('>'):
                    current_id = line[1:].split()[0]
                else:
                    if current_id is not None:
                        compressed = compress_sequence(line)
                        batch.append((current_id, compressed))
                        current_id = None

                        if len(batch) >= batch_size:
                            yield batch
                            batch = []

            if batch:
                yield batch

    except Exception as e:
        logger.warning(f"Failed to process protein FASTA {fasta}: {e}")

def populate_protein_sequences(conn: sqlite3.Connection,
                               protein_fasta_path: str,
                               ):
    logger.info(f"Processing protein FASTA file: {protein_fasta_path}")

    total_sequences = 0

    for batch in stream_all_proteins(protein_fasta_path):
        insert_data = [(seqid, seq_blob) for seqid, seq_blob in batch]
        conn.executemany("""
                         INSERT INTO protein_data (seqID, protein_seq)
                         VALUES (?, ?)
                             ON CONFLICT(seqID) DO UPDATE SET
                             protein_seq = excluded.protein_seq
                         """, insert_data)
        conn.commit()
        total_sequences += len(batch)

    logger.info(f"Processed {total_sequences} protein sequences")

def database_check(db_path: str,
                   output: str,
                   force: bool = False):
    with sqlite3.connect(db_path) as conn:
        cursor = conn.cursor()

        stages = [
            (
                "COG",
                """
                SELECT seqID FROM protein_data
                WHERE COG_ID IS NULL
                """
            ),
            (
                "TMHMM",
                """
                SELECT seqID FROM protein_data
                WHERE no_tmh IS NULL
                """
            ),
            (
                "KEGG",
                """
                SELECT seqID FROM protein_data
                WHERE KEGG_ID IS NULL
                """
            ),
            (
                "Pfam",
                """
                SELECT seqID FROM protein_data
                WHERE Pfam_ID IS NULL
                """
            ),
            (
                "Protein_sequences",
                """
                SELECT seqID FROM protein_data
                WHERE protein_seq IS NULL
                """
            ),
        ]

        for stage, query in stages:
            print(f"Checking for missing data: {stage}")
            cursor.execute(query)
            missing = [row[0] for row in cursor.fetchall()]

            out_file = ensure_path(output, f"missing_{stage}.txt", force=force)
            with open(out_file, 'w') as f:
                for seqid in missing:
                    f.write(f"{seqid}\n")

        genome_stages = [
            (
                "Taxonomy",
                f"""
                SELECT genome_ID FROM genome_data
                WHERE {" OR ".join(f"{col} IS NULL" for col in TAXONOMY_COLUMNS)}
                """
            ),
            (
                "Culture_collection",
                """
                SELECT genome_ID FROM genome_data
                WHERE culture_collection IS NULL
                """
            ),
            (
                "High_level_environment",
                f"""
                SELECT genome_ID FROM genome_data
                WHERE {" OR ".join(f"{col} IS NULL" for col in HIGH_LEVEL_ENV_COLUMNS)}
                """
            ),
            (
                "Low_level_environment",
                f"""
                SELECT genome_ID FROM genome_data
                WHERE {" OR ".join(f"{col} IS NULL" for col in LOW_LEVEL_ENV_COLUMNS)}
                """
            ),
        ]

        for stage, query in genome_stages:
            cursor.execute(query)
            missing = [row[0] for row in cursor.fetchall()]

            out_file = ensure_path(output, f"missing_{stage}.txt", force=force)
            with open(out_file, 'w') as f:
                for genome_id in missing:
                    f.write(f"{genome_id}\n")


def database(cog_gff_tar_path: str,
          kegg_gff_tar_path: str = None,
          pfam_gff_tar_path: str = None,
          tmhmm_tar_path: str = None,
          protein_fasta_path: str = None,
          taxonomy_path: str = None,
          culture_collection_path: str = None,
          high_level_environment_path: str = None,
          low_level_environment_path: str = None,
          output_dir: str = None,
          globdb_version: str = None,
          ):
    """
    Main parsing function - processes COG first with full data, then updates with KEGG and Pfam.
    All annotations use Name= attribute from GFF files.
    """
    logger.info("Processing files: Protein sequences → COG → TMHMM → KEGG → Pfam → Taxonomy → Culture collection "
                "→ High level environment data → Low level environment data → Checking missing data")

    db_path = f"globdb_{globdb_version}_cugo.db"

    if output_dir is None:
        output_dir = '.'

    output_path = Path(output_dir)
    tempdir = output_path / 'tempdir'
    if tempdir.exists():
        shutil.rmtree(tempdir)
    tempdir.mkdir(parents=True, exist_ok=True)

    # Setup database
    conn = setup_database(db_path)

    # ===== STEP 1: Process Protein FASTA file =====
    if protein_fasta_path:
        logger.info("STEP 1: Processing protein sequences...")
        populate_protein_sequences(conn, protein_fasta_path)

    logger.info("Creating seqID index...")
    conn.execute('CREATE INDEX IF NOT EXISTS idx_all_seqid ON protein_data(seqID)')
    conn.commit()
    logger.info("seqID index created")

    # ===== STEP 2: Process COG GFF files =====
    logger.info("STEP 2: Processing COG GFF files...")
    subprocess.run(["tar", "-xzf", cog_gff_tar_path, "-C", str(tempdir)], check=True)

    cog_gff_files = list(tempdir.rglob("*.gff.gz")) + list(tempdir.rglob("*.gff"))
    logger.info(f"Found {len(cog_gff_files)} COG GFF files")

    for gff_file in tqdm(cog_gff_files, desc="Processing COG GFF"):
        gff_data = process_gff_file(str(gff_file))
        if gff_data:
            conn.executemany("""
                INSERT INTO protein_data
                (seqID, parent_ID, aa_length, strand, COG_ID, cugo_number)
                VALUES (?, ?, ?, ?, ?, ?)
                    ON CONFLICT(seqID) DO UPDATE SET
                    parent_ID = excluded.parent_ID,
                    aa_length = excluded.aa_length,
                    strand = excluded.strand,
                    COG_ID = excluded.COG_ID,
                    cugo_number = excluded.cugo_number
            """, gff_data)
            conn.commit()

    logger.info("Creating parent_ID index...")
    conn.execute('CREATE INDEX IF NOT EXISTS idx_parent_id ON protein_data(parent_ID)')
    conn.commit()
    logger.info("parent_ID index created")

    shutil.rmtree(tempdir)
    tempdir.mkdir(parents=True, exist_ok=True)

    # ===== STEP 3: Process TMHMM files =====
    if tmhmm_tar_path:
        logger.info("STEP 3: Processing TMHMM files...")
        subprocess.run(["tar", "-xzf", tmhmm_tar_path, "-C", str(tempdir)], check=True)

        tmhmm_files = list(tempdir.rglob("*_tmhmm_clean"))
        logger.info(f"Found {len(tmhmm_files)} TMHMM files")

        for tmhmm_file in tqdm(tmhmm_files, desc="Processing TMHMM"):
            tmhmm_data = process_tmhmm_file(str(tmhmm_file))
            if tmhmm_data:
                # Changed to handle new entries
                conn.executemany("""
                                 INSERT INTO protein_data (seqID, no_tmh)
                                 VALUES (?, ?)
                                     ON CONFLICT(seqID) DO UPDATE SET
                                     no_tmh = excluded.no_tmh
                                 """, [(seqid, no_tmh) for no_tmh, seqid in tmhmm_data])
                conn.commit()

        shutil.rmtree(tempdir)
        tempdir.mkdir(parents=True, exist_ok=True)

    # ===== STEP 4: Process KEGG GFF files =====
    if kegg_gff_tar_path:
        logger.info("STEP 4: Updating KEGG annotations...")
        subprocess.run(["tar", "-xzf", kegg_gff_tar_path, "-C", str(tempdir)], check=True)

        kegg_gff_files = list(tempdir.rglob("*.gff.gz")) + list(tempdir.rglob("*.gff"))
        logger.info(f"Found {len(kegg_gff_files)} KEGG GFF files")

        for gff_file in tqdm(kegg_gff_files, desc="Updating KEGG IDs"):
            kegg_data = process_gff_file(str(gff_file))
            if kegg_data:
                data = []
                for record in kegg_data:
                    seqID, parent, aa_length, direction, annotation_ID, cugo = record
                    data.append((seqID, parent, aa_length, direction, annotation_ID, cugo, annotation_ID))

                conn.executemany("""
                                 INSERT INTO protein_data (seqID, parent_ID, aa_length, strand, KEGG_ID, cugo_number)
                                 VALUES (?, ?, ?, ?, ?, ?)
                                     ON CONFLICT(seqID) DO UPDATE SET
                                     KEGG_ID = ?
                                 """, data)
                conn.commit()

        shutil.rmtree(tempdir)
        tempdir.mkdir(parents=True, exist_ok=True)

    # ===== STEP 5: Process Pfam GFF files =====
    if pfam_gff_tar_path:
        logger.info("STEP 5: Updating Pfam annotations...")
        subprocess.run(["tar", "-xzf", pfam_gff_tar_path, "-C", str(tempdir)], check=True)

        pfam_gff_files = list(tempdir.rglob("*.gff.gz")) + list(tempdir.rglob("*.gff"))
        logger.info(f"Found {len(pfam_gff_files)} Pfam GFF files")

        for gff_file in tqdm(pfam_gff_files, desc="Updating Pfam IDs"):
            pfam_data = process_gff_file(str(gff_file))
            if pfam_data:
                data = []
                for record in pfam_data:
                    seqID, parent, aa_length, direction, annotation_ID, cugo = record

                    data.append((seqID, parent, aa_length, direction, annotation_ID, cugo, annotation_ID))

                conn.executemany("""
                                 INSERT INTO protein_data (seqID, parent_ID, aa_length, strand, Pfam_ID, cugo_number)
                                 VALUES (?, ?, ?, ?, ?, ?)
                                     ON CONFLICT(seqID) DO UPDATE SET
                                     Pfam_ID = ?
                                 """, data)
                conn.commit()

        shutil.rmtree(tempdir)

    # ===== STEP 6: Process Taxonomy file =====
    if taxonomy_path:
        logger.info("STEP 6: Processing taxonomy data...")
        populate_taxonomy_table(conn, taxonomy_path)

    logger.info("Creating genome_data index...")
    conn.execute('CREATE INDEX IF NOT EXISTS idx_genome_id ON genome_data(genome_ID)')
    conn.commit()
    logger.info("genome_data index created")

    # ==== STEP 7: Process culture collection file ====
    if culture_collection_path:
        logger.info("STEP 7: Processing culture collection data...")
        populate_culture_collection(conn, culture_collection_path)

    # ==== STEP 8: Process high level environment data ====
    if high_level_environment_path:
        logger.info("STEP 8: Processing high level environment data...")
        populate_high_level_environment(conn, high_level_environment_path)

    # ==== STEP 9: Process low level environment data ====
    if low_level_environment_path:
        logger.info("STEP 9: Processing low level environment data...")
        populate_low_level_environment(conn, low_level_environment_path)

    logger.info("Optimizing database with VACUUM...")
    conn.execute("VACUUM")
    conn.commit()
    logger.info("VACUUM complete")

    conn.close()
    logger.info("Database creation complete!")


# ======================================
# METADATA RETRIEVAL
# ======================================
def build_query(include_taxonomy: bool = False,
                include_annotation: bool = False,
                include_culture_collection: bool = False,
                include_high_level_environment: bool = False,
                include_low_level_environment: bool = False,
                batch_size: int = 1) -> str:
    select_cols = [f'p.{col}' for col in BASE_COLUMNS]

    if include_annotation:
        select_cols.extend([f'p.{col}' for col in ANNOTATION_COLUMNS])

    if include_taxonomy:
        select_cols.extend([f'g.{col}' for col in TAXONOMY_COLUMNS])

    if include_culture_collection:
        select_cols.extend([f'g.{col}' for col in CULTURE_COLLECTION_COLUMNS])

    if include_high_level_environment:
        select_cols.extend([f'g.{col}' for col in HIGH_LEVEL_ENV_COLUMNS])

    if include_low_level_environment:
        select_cols.extend([f'g.{col}' for col in LOW_LEVEL_ENV_COLUMNS])

    query = f"SELECT {', '.join(select_cols)} FROM protein_data p"

    needs_genome_join = (
            include_taxonomy or include_culture_collection or
            include_high_level_environment or include_low_level_environment
    )

    if needs_genome_join:
        query += """
            LEFT JOIN genome_data g ON 
            CASE 
                WHEN instr(p.seqID, '___') > 0 
                THEN substr(p.seqID, 1, instr(p.seqID, '___') - 1)
                ELSE p.seqID
            END = g.genome_ID
            """

    if batch_size > 1:
        placeholders = ','.join(['?' for _ in range(batch_size)])
        query += f" WHERE p.seqID IN ({placeholders})"
    else:
        query += " WHERE p.seqID = ?"

    return query


def get_header(include_taxonomy: bool,
               include_annotation: bool,
               include_culture_collection: bool,
               include_high_level_environment: bool,
               include_low_level_environment: bool):

    header = BASE_COLUMNS.copy()

    if include_annotation:
        header.extend(ANNOTATION_COLUMNS)

    if include_taxonomy:
        header.extend(TAXONOMY_COLUMNS)

    if include_culture_collection:
        header.extend(CULTURE_COLLECTION_COLUMNS)

    if include_high_level_environment:
        header.extend(HIGH_LEVEL_ENV_COLUMNS)

    if include_low_level_environment:
        header.extend(LOW_LEVEL_ENV_COLUMNS)

    return header


def process_batch(db_path: str,
                  batch: list,
                  include_taxonomy: bool,
                  include_annotation: bool,
                  include_culture_collection: bool,
                  include_high_level_environment: bool,
                  include_low_level_environment: bool):

    conn = sqlite3.connect(db_path)
    cursor = conn.cursor()

    query = build_query(
        include_taxonomy=include_taxonomy,
        include_annotation=include_annotation,
        include_culture_collection=include_culture_collection,
        include_high_level_environment=include_high_level_environment,
        include_low_level_environment=include_low_level_environment,
        batch_size=len(batch)
    )

    cursor.execute(query, batch)
    results = cursor.fetchall()

    result_dict = {row[0]: row for row in results}

    conn.close()
    return batch, result_dict


def meta(db_path: str,
         fasta: str,
         output: str,
         threads: int = 1,
         include_annotation: bool = False,
         include_taxonomy: bool = False,
         include_culture_collection: bool = False,
         include_high_level_environment: bool = False,
         include_low_level_environment: bool = False,
         all_metadata: bool = False,
         force: bool = False):

    if fasta:
        seq_dict = read_fasta_to_dict(fasta)
        seq_ids = list(seq_dict.keys())
        dataset_name = determine_dataset_name(fasta, '.', 0)
        output_path = ensure_path(
            output,
            f'{dataset_name}_metadata.tsv',
            force=force
        )
    else:
        logger.error(
            "No valid input file found. Please specify path to protein FASTA"
        )
        return

    # Enable everything if --all-metadata is set
    if all_metadata:
        include_annotation = True
        include_taxonomy = True
        include_culture_collection = True
        include_high_level_environment = True
        include_low_level_environment = True

    # Ensure at least one metadata type is requested
    if not any([
        include_annotation,
        include_taxonomy,
        include_culture_collection,
        include_high_level_environment,
        include_low_level_environment
    ]):
        logger.error(
            "Must specify at least one metadata type "
            "(--taxonomy, --annotation, --culture-collection, "
            "--high-level-environment, --low-level-environment, or --all-metadata)"
        )
        return

    BATCH_SIZE = 500

    header = get_header(
        include_taxonomy,
        include_annotation,
        include_culture_collection,
        include_high_level_environment,
        include_low_level_environment
    )

    # Split seq_ids into batches
    batches = [
        seq_ids[i:i + BATCH_SIZE]
        for i in range(0, len(seq_ids), BATCH_SIZE)
    ]

    found_ids = set()
    missing_ids = set()
    batch_results = {}

    # Process batches in parallel
    with ThreadPoolExecutor(max_workers=threads) as executor:
        futures = {
            executor.submit(
                process_batch,
                db_path,
                batch,
                include_taxonomy,
                include_annotation,
                include_culture_collection,
                include_high_level_environment,
                include_low_level_environment
            ): idx
            for idx, batch in enumerate(batches)
        }

        for future in tqdm(
                as_completed(futures),
                total=len(batches),
                desc="Querying database"
        ):
            batch_idx = futures[future]
            try:
                batch, result_dict = future.result()
                batch_results[batch_idx] = (batch, result_dict)
            except Exception as e:
                logger.error(
                    f"Error processing batch {batch_idx}: {e}"
                )

    # Write results in original order
    with open(output_path, 'w') as out:
        out.write('\t'.join(header) + '\n')

        for batch_idx in sorted(batch_results.keys()):
            batch, result_dict = batch_results[batch_idx]

            for seq_id in batch:
                if seq_id in result_dict:
                    found_ids.add(seq_id)
                    row = [
                        str(x) if x is not None else ''
                        for x in result_dict[seq_id]
                    ]
                    out.write('\t'.join(row) + '\n')
                else:
                    missing_ids.add(seq_id)

    logger.info(f"Metadata has been saved to {output_path}")

    if missing_ids:
        logger.warning(
            f"{len(missing_ids)} sequence IDs were not found in the database"
        )



def list_metadata():
    print("───────────────────────────")
    print("Protein metadata")
    print("───────────────────────────")
    print("  • sequence information")
    for item in BASE_COLUMNS[1:]:
        print(f"      - {item}")
    print()
    print("  • annotations")
    for item in ANNOTATION_COLUMNS:
        print(f"      - {item}")
    print()  # blank line

    print("───────────────────────────")
    print("Genome metadata")
    print("───────────────────────────")
    print("  • taxonomy")
    for item in TAXONOMY_COLUMNS:
        print(f"      - {item}")
    print()

    print("  • culture information")
    for item in CULTURE_COLLECTION_COLUMNS:
        print(f"      - {item}")
    print()

    print("  • high-level environment data")
    for item in HIGH_LEVEL_ENV_COLUMNS:
        print(f"      - {item}")
    print()

    print("  • low-level environment data")
    for item in LOW_LEVEL_ENV_COLUMNS:
        print(f"      - {item}")
    print()

def stream_sequence_ids(db_path, batch_size):
    conn = sqlite3.connect(db_path)
    cursor = conn.cursor()

    cursor.execute("SELECT seqID FROM protein_data")

    batch = []
    for row in cursor:
        batch.append(row[0])
        if len(batch) >= batch_size:
            yield batch
            batch = []
    if batch:
        yield batch

    conn.close()

def fetch_sequences(db_path, seq_ids):
    conn = sqlite3.connect(db_path)
    cursor = conn.cursor()

    placeholders = ', '.join(['?'] * len(seq_ids))
    cursor.execute(f"""SELECT seqID, protein_seq FROM protein_data WHERE seqID in ({placeholders})""", seq_ids)

    records = []
    for seq_id, blob in cursor.fetchall():
        if blob:
            records.append((seq_id, decompress_sequence(blob)))

    return records



def export_fasta(db_path: str,
                  output: str,
                  threads: int = 1,
                  force: bool =False):
    protein_fasta_file = ensure_path(output, 'globdb_all_prot.faa', force=force)

    batch_size = 900

    with open(protein_fasta_file, 'w') as out:
        with ThreadPoolExecutor(max_workers=threads) as executor:
            futures = set()

            for batch in stream_sequence_ids(db_path, batch_size):
                futures.add(executor.submit(fetch_sequences, db_path, batch))

                if len(futures) >= threads:
                    done = next(as_completed(futures))
                    futures.remove(done)

                    for seqID, seq in done.result():
                        out.write(f">{seqID}\n{seq}\n")

            for future in as_completed(futures):
                for seqID, seq in future.result():
                    out.write(f">{seqID}\n{seq}\n")




from .util import *
from .database import BASE_COLUMNS, ANNOTATION_COLUMNS, TAXONOMY_COLUMNS, CULTURE_COLLECTION_COLUMNS, HIGH_LEVEL_ENV_COLUMNS, LOW_LEVEL_ENV_COLUMNS

import logging
import numpy as np
import pandas as pd
import openTSNE
from sklearn.cluster import DBSCAN
import matplotlib.pyplot as plt
import sqlite3
import json
import hashlib


logger = logging.getLogger(__name__)

CASM_BLAST_OUTPUT_COLUMNS = ["qseqid", "sseqid", "score"]

def build_alignment_matrix_split(align_file: str,
                                 output: str = None,
                                 force: bool = False):
    """
    Builds alignment matrix from input Blast Tabular Output to be used in tSNE embedding

    Args:
        align_file (str): Path to Blast Tabular Output
        output (str): Path to desired output directory
        force (bool): If set, existing files will be overwritten

    Returns:
        matrix: Alignment matrix as dense numpy matrix
        queries (list): List of query protein IDs
        targets (list): List of reference protein IDs
        matrix_file (str): Path to .npy file containing numpy matrix
        metadata_file (str): Path to .json file containing matrix metadata (queries, targets, matrix stats)
    """
    logger.info(f"Building alignment matrix from: {align_file} (split parsing)")

    # ===============================
    # Storage preparation
    # ===============================

    # set up empty sets for storing query and reference protein IDs
    queries_set = set()
    targets_set = set()

    # =======================================================
    # First pass over alignment file: Retrieve protein IDs
    # =======================================================
    with open(align_file, 'r') as f:
        for line_num, line in enumerate(f, 1):
            line = line.rstrip('\n\r')
            if not line:
                continue

            try:
                # split tab separated line to get query ID, target ID and score
                parts = line.split("\t")
                if len(parts) != 3:
                    logger.warning(f"Line {line_num}: Expected 3 fields, got {len(parts)}. Skipping.")
                    continue
                query, target, score_str = parts

                # add protein IDs to respective sets
                queries_set.add(query)
                targets_set.add(target)
            except Exception as e:
                logger.warning(f"Line {line_num}: Error parsing line. Skipping. ({e})")
                continue

    # sort both sets
    queries = sorted(queries_set)
    targets = sorted(targets_set)

    # determine indices for each ID for matrix construction and save {protID: index}
    query_to_idx = {q: i for i, q in enumerate(queries)}
    target_to_idx = {t: i for i, t in enumerate(targets)}

    # delete ID sets for efficient memory handling
    del queries_set, targets_set

    # set up empty numpy matrix with correct dimensions
    matrix = np.zeros((len(queries), len(targets)), dtype=np.float32)

    # =======================================================
    # Second pass over alignment file: Matrix construction
    # =======================================================
    with open(align_file, 'r') as f:
        for line_num, line in enumerate(f, 1):
            line = line.rstrip('\n\r')
            if not line:
                continue

            # split tab separated line to get query ID, target ID and score
            try:
                parts = line.split("\t")
                if len(parts) != 3:
                    continue
                query, target, score_str = parts

                score = float(score_str)
                i = query_to_idx[query]
                j = target_to_idx[target]

                # assign alignment score to correct matrix cell
                matrix[i, j] = score
            except ValueError as e:
                logger.warning(f"Line {line_num}: Invalid score '{score_str}'. Skipping. ({e})")
                continue
            except KeyError as e:
                logger.warning(f"Line {line_num}: Key not found in mapping. Skipping. ({e})")
                continue
            except Exception as e:
                logger.warning(f"Line {line_num}: Unexpected error. Skipping. ({e})")
                continue

    # ===============================
    # Determine matrix statistics
    # ===============================
    non_zero_elements = np.count_nonzero(matrix)
    matrix_elements = len(queries) * len(targets)
    sparsity = (matrix_elements - non_zero_elements) / matrix_elements * 100

    logger.info(f"Matrix construction completed")
    logger.info(f"Non-zero elements: {non_zero_elements:,} ({100 - sparsity:.2f}% filled)")
    logger.info(f"Matrix sparsity: {sparsity:.2f}%")
    logger.info(f"Score range: {matrix.min():.2f} - {matrix.max():.2f}")

    dataset_name = determine_dataset_name(align_file, '.', 0)

    # =============================================
    # Matrix and matrix metadate output to files
    # =============================================
    if output:
        matrix_file = ensure_path(output, f"{dataset_name}_matrix.npy", force=force)
        metadata_file = ensure_path(output, f"{dataset_name}_matrix_metadata.json", force=force)
    else:
        matrix_file = ensure_path(target=f"{dataset_name}_matrix.npy", force=force)
        metadata_file = ensure_path(target=f"{dataset_name}_matrix_metadata.json", force=force)

    np.save(matrix_file, matrix)
    logger.info(f"Matrix saved to: {matrix_file}")

    metadata = {
        "queries": queries,
        "targets": targets,
        "matrix_shape": matrix.shape,
        "non_zero_elements": int(non_zero_elements),
        "sparsity": float(sparsity),
        "score_range": [float(matrix.min()), float(matrix.max())]
    }

    with open(metadata_file, 'w') as file:
        json.dump(metadata, file, indent=2)
    logger.info(f"Matrix metadata saved to: {metadata_file}")

    return matrix, queries, targets, matrix_file, metadata_file

def load_alignment_matrix_from_file(matrix_path: str,
                                    metadata_path: str):
    """
    Loads alignment matrix and matrix metadata into memory

    Args:
        matrix_path (str): Path to .npy file containing alignment matrix
        metadata_path (str): Path to matrix metadata file

    Returns:
        matrix (numpy array): Alignment matrix
        queries (list): List of query protein IDs
        targets (list): List of reference protein IDs
    """
    matrix = np.load(matrix_path)

    with open(metadata_path, "r") as file:
        metadata = json.load(file)

    queries = metadata["queries"]
    targets = metadata["targets"]

    logger.info(f"Loaded matrix shape: {matrix.shape}")
    logger.info(f"Loaded {len(queries)} queries and {len(targets)} targets")

    return matrix, queries, targets

def create_embedding_file(output_file: str,
                          embedding: np.ndarray,
                          queries: list,
                          clusters: np.ndarray,
                          col_names: list,
                          ):
    """
    Creates a TSV file containing protein embeddings with metadata and cluster assignments.

    Args:
        output_file (str): Path to the output TSV file to be created
        embedding (np.ndarray): 2D array of embedding vectors, shape (n_proteins, n_dimensions)
        queries (list): List of protein IDs in format "genomeID_locusNr" (e.g., "genome_ABC_001")
        clusters (np.ndarray): 1D array of cluster assignments for each protein
        col_names (list): Column names for each embedding dimension (e.g., ["dim_0", "dim_1", ...])
    """
    logger.info(f"Embedded {len(queries)} proteins")

    loci = [seqID.rsplit("_", 1)[1] for seqID in queries]
    genome_ids = [seqID.rsplit("_", 1)[0] for seqID in queries]

    with open(output_file, "w", encoding="utf-8") as f:
        base_cols = ["seqID", "locus_nr", "genome_ID"] + col_names + ["cluster"]

        f.write("\t".join(base_cols) + "\n")
        for q, c, l, g, emb in zip(queries, clusters, loci, genome_ids, embedding):
            row = [q, l, g] + [f"{x:.6f}" for x in emb] + [str(c)]
            f.write("\t".join(row) + "\n")

def create_embedding_dataframe(embedding: np.ndarray,
                               queries: list,
                               clusters: np.ndarray,
                               col_names: list
                               ):
    """
    Creates DataFrame from t-SNE embedding results with clustering and optional metadata.

    Args:
        embedding (np.ndarray): t-SNE embedding coordinates
        queries (list): List of protein IDs corresponding to embedding rows
        clusters (np.ndarray): Cluster assignments for each protein
        col_names (list): Column names for embedding dimensions

    Returns:
        df (pd.DataFrame): DataFrame with protein IDs, coordinates, clusters, and metadata
    """
    logger.debug(f"Creating embedding dataframe with {len(queries)} proteins")
    df = pd.DataFrame(embedding, columns=col_names)
    df['seqID'] = queries
    df['cluster'] = clusters

    logger.debug("Extracting locus and genome information from protein IDs")

    # extract locus and genome info from protein IDs
    df['locus_nr'] = df['seqID'].astype(str).str.rsplit("_", n=1).str[1]
    df['genome_ID'] = df['seqID'].astype(str).str.rsplit("_", n=1).str[0]

    # reorder columns
    base_cols = ['seqID', 'locus_nr', 'genome_ID'] + col_names + ['cluster']
    df = df[base_cols]

    logger.debug(f"Final dataframe shape: {df.shape}")
    return df


def tsne_embedding(matrix: np.ndarray,
                   queries: list,
                   output: str,
                   basename: str,
                   perplexity: int = 50,
                   iterations: int = 500,
                   exaggeration: int = 6,
                   threads: int = 1,
                   force: bool = False
                   ):
    """
    Performs t-SNE embedding and DBSCAN clustering on alignment matrix.

    Creates early (exaggerated) and final t-SNE embeddings with DBSCAN clustering
    applied to the early embedding. Same cluster labels are used for both embeddings.

    Args:
        matrix (np.ndarray): Alignment matrix for embedding
        queries (list): List of protein IDs corresponding to matrix rows
        output (str): Output directory (default: current working directory)
        basename (str): Base name for output files
        perplexity (int): t-SNE perplexity parameter
        iterations (int): Number of optimization iterations per phase
        exaggeration (int): Early exaggeration factor
        threads (int): Number of threads for parallel processing
        force (bool): Overwrite existing files if True

    Returns:
        early_filename: Path to early embedding TSV file
        final_filename: Path to final embedding TSV file
    """
    logger.info("Starting t-SNE embedding process")
    logger.info(f"Matrix shape: {matrix.shape}")
    logger.info(f"Perplexity: {perplexity}, Iterations: {iterations}, Exaggeration: {exaggeration}")
    logger.info(f"Using {threads} threads")

    # ========================
    # Multiscale affinites
    # ========================

    # compute affinity matrix using multiple perplexity values for local and global structure
    logger.info("Computing multiscale affinities")
    affinities = openTSNE.affinity.Multiscale(matrix,
                                              perplexities=[20, perplexity],
                                              metric="cosine",
                                              n_jobs=threads)

    # ========================
    # Initialize embedding
    # ========================

    # initialize with PCA for better convergence
    logger.info("Initializing t-SNE with PCA")
    pca_init = openTSNE.initialization.pca(matrix)

    # create t-SNE embedding with auto learning rate and single DOF
    tsne_embed = openTSNE.TSNEEmbedding(pca_init,
                                        affinities,
                                        n_jobs=threads,
                                        dof=1,
                                        learning_rate="auto")


    # ============================
    # Early optimization phase
    # ============================

    # optimize with exaggeration to spread clusters; lower momentum for exploration
    logger.info(f"Starting early optimization with exaggeration={exaggeration}")
    tsne_embed.optimize(n_iter=iterations,
                        exaggeration=exaggeration,
                        momentum=0.5,
                        inplace=True,
                        n_jobs=threads)
    logger.info("Early optimization completed")

    # =================================
    # Clustering on early embedding
    # =================================

    # apply DBSCAN to identify protein clusters
    logger.info("Performing DBSCAN clustering on early embedding")
    clustering = DBSCAN(eps=1, min_samples=10).fit(tsne_embed)

    # calculate cluster statistics (label -1 indicates noise points)
    n_clusters = len(set(clustering.labels_)) - (1 if -1 in clustering.labels_ else 0)
    n_noise = list(clustering.labels_).count(-1)
    logger.info(f"DBSCAN clustering results: {n_clusters} clusters, {n_noise} noise points")

    # ================================
    # Save early embedding results
    # ================================
    logger.debug("Saving early embedding results")
    early_filename = ensure_path(output, f"{basename}_tsne_early_clust.tsv", force=force)
    create_embedding_file(early_filename, tsne_embed, queries, clustering.labels_, ['tsne1', 'tsne2'])
    logger.info(f"Early embedding saved to: {early_filename}")

    # ========================
    # Final optimization phase
    # ========================

    # fine-tune embedding without exaggeration; higher momentum for stability
    logger.info("Starting final optimization")
    tsne_embed.optimize(n_iter=iterations,
                        momentum=0.8,
                        inplace=True,
                        n_jobs=threads)
    logger.info("Final optimization completed")

    # ================================
    # Save final embedding results
    # ===============================
    logger.debug("Saving final embedding results")
    final_filename = ensure_path(output, f"{basename}_tsne_final_clust.tsv", force=force)
    create_embedding_file(final_filename, tsne_embed, queries, clustering.labels_, ['tsne1', 'tsne2'])
    logger.info(f"Final embedding saved to: {final_filename}")

    return early_filename, final_filename

def fetch_protein_metadata(db_path: str,
                           protein_ids: list,
                           column: str,
                           batch_size: int = 500
                           ):
    valid_protein_cols = BASE_COLUMNS[1:] + ANNOTATION_COLUMNS
    valid_genome_cols = TAXONOMY_COLUMNS + CULTURE_COLLECTION_COLUMNS + LOW_LEVEL_ENV_COLUMNS + HIGH_LEVEL_ENV_COLUMNS

    if column not in valid_protein_cols and column not in valid_genome_cols:
        logger.warning(f"No valid protein metadata columns requested")
        return pd.DataFrame({'seqID': protein_ids})

    conn = sqlite3.connect(db_path)
    all_results = []
    n_batches = (len(protein_ids) + batch_size - 1) // batch_size
    logger.info(f"Fetching metadata in {n_batches} batches")

    for i in range(0, len(protein_ids), batch_size):
        batch = protein_ids[i:i + batch_size]
        placeholders = ','.join(['?'] * len(batch))

        if column in valid_protein_cols:
            query = f"""SELECT seqID, {column} FROM protein_data WHERE seqID IN ({placeholders})"""
            batch_df = pd.read_sql_query(query, conn, params=batch)

        elif column in valid_genome_cols:
            query = f"""
                SELECT p.seqID, g.{column}
                FROM protein_data p
                LEFT JOIN genome_data g ON 
                    CASE 
                        WHEN instr(p.seqID, '___') > 0 
                        THEN substr(p.seqID, 1, instr(p.seqID, '___') - 1)
                        ELSE p.seqID
                    END = g.genome_ID
                WHERE p.seqID IN ({placeholders})
            """
            batch_df = pd.read_sql_query(query, conn, params=batch)

        all_results.append(batch_df)

    conn.close()
    df = pd.concat(all_results, ignore_index=True)
    logger.info(f"Retrieved {len(df)} metadata records total")
    return df

def extend_embedding_with_metadata(df: pd.DataFrame,
                                   db_path: str,
                                   protein_col: str = None,
                                   ):
    if not db_path:
        logger.warning("No database path provided, skipping metadata enrichment")
        return df

    if protein_col is not None:
        logger.info(f"Fetching protein metadata: {protein_col}")
        protein_metadata = fetch_protein_metadata(db_path, df['seqID'].tolist(), protein_col)
        df = pd.merge(df, protein_metadata, on='seqID', how='left')


    return df


def plot_clusters(tsv_file: str,
                  output: str,
                  db_path: str = None,
                  metadata_protein: str = None,
                  svg: bool = False,
                  force: bool = False,
                  show_cluster_numbers: bool = False
                  ):
    """
    Create a t-SNE scatter plot visualization of DBSCAN clustering results.

    Reads clustering results from a TSV file and generates a scatter plot showing
    the t-SNE embedding with different colors for each cluster. Noise points
    (cluster -1) are displayed in light gray with reduced opacity.

    Args:
        tsv_file (str): Path to input TSV file containing tSNE coordinates and cluster assignment
        output (str): Base path for output plot file (adds "_tsne_clusters" suffix)
        svg (bool): Generate plot in SVG format
        force (bool): Overwrite existing files if True
        show_cluster_numbers (bool): Display cluster number on cluster centers in output plot
    """
    # load clustering results from TSV file
    logger.info(f"Creating t-SNE plot from: {tsv_file}")
    df = pd.read_csv(tsv_file, sep='\t')
    logger.info(f"Loaded {len(df)} data points for plotting")

    if 'seqID' not in df.columns:
        raise KeyError("Expected column 'seqID' not found in TSV")

    df['seqID'] = df['seqID'].astype(str).str.strip()
    df['genome_ID'] = (
        df['seqID']
        .apply(parse_protein_identifier)
        .astype(str)
        .str.strip()
    )

    color_column = None
    if db_path:
        if metadata_protein:
            logger.info(f"Fetching metadata column '{metadata_protein}' for coloring")
            df = extend_embedding_with_metadata(df, db_path, protein_col=metadata_protein)
            color_column = metadata_protein

    if metadata_protein  and not db_path:
        logger.info("Metadata selected but path to SQLite DB not provided, coloring by cluster")
        color_column = 'cluster'
    elif color_column is None or color_column not in df.columns:
        logger.info("No metadata column selected, coloring by cluster")
        color_column = 'cluster'

    # ========================
    # Initialize plot
    # ========================
    plt.figure(figsize=(14, 10))

    if metadata_protein == 'culture_collection' and 'culture_collection' in df.columns:
        overlay_mask = df['culture_collection'] == 1
        non_overlay_mask = ~overlay_mask

        plt.scatter(df.loc[non_overlay_mask, 'tsne1'],
                    df.loc[non_overlay_mask, 'tsne2'],
                    color='lightgray',
                    s=15,
                    alpha=0.5,
                    edgecolors='none')

        if overlay_mask.any():
            n_overlay = overlay_mask.sum()
            logger.info(f"Overlaying {n_overlay} points with culture_collection == 1")
            plt.scatter(df.loc[overlay_mask, 'tsne1'],
                        df.loc[overlay_mask, 'tsne2'],
                        c='red',
                        s=25,
                        alpha=0.6,
                        edgecolors='black',
                        linewidths=0.5,
                        marker='o',
                        label='Culture Collection',
                        zorder=10)


    elif color_column == 'cluster':
        # create masks to separate noise points (cluster -1) from clustered points
        noise_mask = df['cluster'] == -1
        clustered_mask = df['cluster'] != -1

        n_noise = noise_mask.sum()
        n_clustered = clustered_mask.sum()
        logger.info(f"Plotting {n_clustered} clustered points and {n_noise} noise points")

        # ========================
        # Plot noise points
        # ========================

        # display unclustered points in light gray with low opacity
        if noise_mask.any():
            plt.scatter(df.loc[noise_mask, 'tsne1'],
                        df.loc[noise_mask, 'tsne2'],
                        c='lightgray',
                        alpha=0.4,
                        s=12,
                        edgecolors='none',
                        label='Noise')

        # ========================
        # Plot clustered points
        # ========================
        if clustered_mask.any():
            unique_clusters = df.loc[clustered_mask, 'cluster'].unique()

            # use tab20 colormap for up to 20 clusters, switch to hsv for more
            colors = plt.cm.tab20(np.linspace(0, 1, len(unique_clusters)))
            if len(unique_clusters) > 20:
                colors = plt.cm.hsv(np.linspace(0, 1, len(unique_clusters)))

            logger.info(f"Plotting {len(unique_clusters)} unique clusters")

            # plot each cluster with unique color
            for i, cluster in enumerate(unique_clusters):
                cluster_mask = df['cluster'] == cluster
                cluster_size = cluster_mask.sum()
                logger.debug(f"Cluster {cluster}: {cluster_size} points")
                plt.scatter(df.loc[cluster_mask, 'tsne1'],
                            df.loc[cluster_mask, 'tsne2'],
                            c=[colors[i]],
                            s=15,
                            alpha=0.7,
                            edgecolors='white',
                            linewidths=0.3,
                            label=f'Cluster {cluster}')

    else:
        unique_vals = df[color_column].dropna().unique()

        color_map = {
            metadatum: '#' + hashlib.md5(str(metadatum).encode()).hexdigest()[:6]
            for metadatum in unique_vals
        }

        na_mask = df[color_column].isna()
        if na_mask.any():
            plt.scatter(df.loc[na_mask, 'tsne1'], df.loc[na_mask, 'tsne2'],
                        c='lightgray', alpha=0.4, s=12, edgecolors='none', label='No data')

        if pd.api.types.is_string_dtype(df[color_column]) or pd.api.types.is_categorical_dtype(df[color_column]):
            for val in unique_vals:
                mask = df[color_column] == val
                plt.scatter(df.loc[mask, 'tsne1'], df.loc[mask, 'tsne2'],
                            color=color_map[val], s=15, alpha=0.7,
                            edgecolors='white', linewidths=0.3, label=str(val))
        else:
            plt.scatter(df['tsne1'], df['tsne2'],
                        c=df[color_column], cmap='viridis', s=15,
                        alpha=0.7, edgecolors='white', linewidths=0.3)


    if show_cluster_numbers and 'cluster' in df.columns:
        clustered_mask = df['cluster'] != -1
        if clustered_mask.any():
            unique_clusters = df.loc[clustered_mask, 'cluster'].unique()
            for cluster in unique_clusters:
                cluster_mask = df['cluster'] == cluster
                centroid_x = df.loc[cluster_mask, 'tsne1'].mean()
                centroid_y = df.loc[cluster_mask, 'tsne2'].mean()
                plt.annotate(str(cluster),
                             xy=(centroid_x, centroid_y),
                             fontsize=16,
                             fontweight='bold',
                             ha='center',
                             va='center',
                             color='black',
                             alpha=1,
                             zorder=100)

    # =============================
    # Configure plot appearance
    # =============================
    plt.xlabel('t-SNE 1', fontsize=12)
    plt.ylabel('t-SNE 2', fontsize=12)
    plt.title('t-SNE Embedding with DBSCAN Clusters', fontsize=14)

    if color_column == 'cluster' and 'cluster' in df.columns:
        unique_clusters = df[df['cluster'] != -1]['cluster'].unique()
        if len(unique_clusters) <= 60:
            plt.legend(bbox_to_anchor=(1.05, 1), loc='upper left', fontsize=10)
        else:
            logger.info("Cluster number exceeds 60, omitting legend.")
    elif color_column != 'cluster':
        unique_vals = df[color_column].dropna().unique()
        if len(unique_vals) <= 60:
            plt.legend(bbox_to_anchor=(1.05, 1), loc='upper left', fontsize=10)
        else:
            logger.info("Cluster number exceeds 60, omitting legend.")

    plt.tight_layout()


    # =============================
    # Determine output filename
    # =============================

    # parse input filename to determine if early or final embedding
    if 'early_clust' in tsv_file:
        prefix = determine_dataset_name(tsv_file, '.', 0, '_tsne_early_clust')
        base_name = f"{prefix}_tsne_early"
    elif 'final_clust' in tsv_file:
        prefix = determine_dataset_name(tsv_file, '.', 0, '_tsne_final_clust')
        base_name = f"{prefix}_tsne_final"
    else:
        raise ValueError(f"Unexpected TSV filename: {tsv_file}")

    if color_column and color_column != 'cluster':
        filename = f"{base_name}_by_{color_column}.{ 'svg' if svg else 'png' }"
    else:
        filename = f"{base_name}_clusters.{ 'svg' if svg else 'png' }"


    # ========================
    # Save plot
    # ========================
    output_file = ensure_path(output, filename, force=force)
    plt.savefig(output_file, dpi=300, bbox_inches='tight')
    logger.info(f"Plot saved to: {output_file}")

    return output_file


def casm_select(final_embedding_file: str,
         fasta: str,
         no_cluster: int,
         output: str,
         force: bool = False):
    """
    Extract proteins belonging to a specific cluster into a separate FASTA file.

    Reads the t-SNE clustering results and extracts all protein sequences that
    belong to the specified cluster number, writing them to a new FASTA file.

    Args:
        final_embedding_file (str): Path to TSV file with t-SNE embeddings and cluster assignments
        fasta (str): Path to input FASTA file containing all protein sequences
        no_cluster (int): Cluster number to extract
        output (str): Output directory path
        force (bool): Overwrite existing files if True

    Returns:
        str: Path to the output FASTA file containing selected cluster sequences
    """
    # ===========================
    # Prepare output filename
    # ===========================

    # generate output filename based on input file and cluster number
    prefix = determine_dataset_name(final_embedding_file, '.', 0, '_tsne_final_clust')
    cluster_fasta = ensure_path(output, f"{prefix}_cluster_{no_cluster}.faa", force=force)

    # ============================================
    # Extract protein IDs from target cluster
    # ============================================
    cluster_seqIDs = []
    with open(final_embedding_file, 'r') as file:
        # parse header to find column indices
        header = file.readline().strip().split()
        cluster_idx = header.index("cluster")
        seqIDx = header.index("seqID")

        # iterate through data rows and collect protein IDs matching target cluster
        for line in file:
            parts = line.strip().split('\t')
            seqID = parts[seqIDx]
            prot_cluster_no = parts[cluster_idx]

            if int(prot_cluster_no) == int(no_cluster):
                cluster_seqIDs.append(seqID)

    # ==================================
    # Load original FASTA sequences
    # ==================================
    input_fasta_dict = read_fasta_to_dict(fasta)

    # ============================================
    # Write cluster sequences to output FASTA
    # ============================================

    # write sequences for all proteins in the selected cluster
    with open(cluster_fasta, 'w') as file:
        for id in cluster_seqIDs:
            if id in input_fasta_dict:
                file.write(f">{id}\n{input_fasta_dict[id]}\n")

    return cluster_fasta

# ==================================================
#
#                   CLI FUNCTIONS
#
# ==================================================

def matrix(fasta: str,
          output: str,
          subset: str,
          subset_size: int,
          threads: int,
          force: bool = False
          ):
    """
    Build alignment matrix from FASTA sequences using DIAMOND alignment.

    Creates a subset of sequences (either provided or randomly sampled), performs
    DIAMOND alignment against the full dataset, and constructs an alignment matrix
    from the results. The matrix represents alignment scores between sequences.

    Args:
        fasta (str): Path to input FASTA file with protein sequences
        output (str): Base name for output files
        subset (str): Path to subset FASTA file (if None, creates random subset)
        subset_size (int): Number of sequences for subset (ignored if subset provided)
        threads (int): Number of threads for DIAMOND alignment
        force (bool): Overwrite existing files if True

    Returns:
        matrix_file (str): Path to .npy file containing numpy matrix
        metadata_file (str): Path to .json file containing matrix metadata (queries, targets, matrix stats)
    """
    logger.info("=== Starting Matrix Construction ===")
    logger.info(f"Input FASTA: {fasta}")
    logger.info(f"Output basename: {output}")
    logger.info(f"Threads: {threads}")
    if subset:
       logger.info(f"Using provided subset: {subset}")
       subset_fasta = subset
       subset_size = count_fasta_sequences(subset_fasta)
    else:
       logger.info("Creating random subset from input FASTA")
       subset_fasta = fasta_subsample(fasta, output, subset_size, force=force)

    logger.info("=== Phase 1: DIAMOND Alignment ===")
    align_output = run_diamond_alignment(fasta, subset_fasta, subset_size, threads, CASM_BLAST_OUTPUT_COLUMNS, force=force)

    logger.info("=== Phase 2: Matrix Construction ===")
    _, _, _, matrix_file, metadata_file = build_alignment_matrix_split(align_output, output, force)

    return matrix_file, metadata_file


def cluster(matrix_path: str,
         matrix_metadata_path: str,
         output: str,
         perplexity: int = 50,
         iterations: int = 500,
         exaggeration: int = 6,
         threads: int = 1,
         force: bool = False,
         ):
    """
    Perform t-SNE embedding and DBSCAN clustering on alignment matrix.

    Creates early (exaggerated) and final t-SNE embeddings with DBSCAN clustering
    applied to the early embedding. Same cluster labels are used for both embeddings.
    Optionally enriches results with protein and genome metadata.

    Args:
        matrix_path (str): Path to alignment matrix file
        matrix_metadata_path (str): Path to matrix metadata file
        output (str): Base name for output files
        perplexity (int): t-SNE perplexity parameter
        iterations (int): Number of optimization iterations per phase
        exaggeration (int): Early exaggeration factor
        threads (int): Number of threads for parallel processing
        force (bool): Overwrite existing files if True

    Returns:
        early_filename (str), final_filename (str): Paths to early and final embedding TSV files
    """
    logger.info("=== Starting t-SNE Embedding ===")
    logger.info(f"Matrix file: {matrix_path}")
    logger.info(f"Metadata file: {matrix_metadata_path}")
    if output:
       logger.info(f"Output basename: {output}")

    prefix = determine_dataset_name(matrix_path, '.', 0, '_matrix')

    matrix, queries, targets = load_alignment_matrix_from_file(matrix_path, matrix_metadata_path)

    early_filename, final_filename = tsne_embedding(matrix=matrix,
                              queries=queries,
                              output=output,
                              basename=prefix,
                              perplexity=perplexity,
                              iterations=iterations,
                              exaggeration=exaggeration,
                              threads=threads,
                              force=force,
                              )
    logger.info("=== t-SNE Embedding Completed ===")
    return early_filename, final_filename


def casm_plot(clust_path: str,
              output: str,
              metadata_protein: str,
              db_path: str,
              svg: bool = False,
              force: bool = False,
              show_cluster_numbers: bool = False):
    """
    Generate t-SNE cluster visualization plots for early and final embeddings.

    Creates scatter plots showing DBSCAN clustering results for both early
    (exaggerated) and final t-SNE embeddings. Generates two plot files with
    "_early_tsne_clusters" and "_final_tsne_clusters" suffixes.

    Args:
        clust_path (str): Path to  clustering results TSV file
        output (str): Base name for output plot files

    Returns:
        early_cluster_plot
        full_cluster_plot
    """
    logger.info('=== Starting Plot Generation ===')
    logger.info(f'Plotting file: {clust_path}')
    logger.info(f'Output basename: {output}')

    # Generate plots
    cluster_plot = plot_clusters(clust_path,
                                 output=output,
                                 metadata_protein=metadata_protein,
                                 db_path=db_path,
                                 force=force,
                                 svg=svg,
                                 show_cluster_numbers=show_cluster_numbers)

    logger.info("=== Plot Generation Completed ===")

    return cluster_plot


def casm(fasta: str,
         output: str,
         subset: str,
         subset_size: int,
         db_path: str = None,
         threads: int = 1,
         perplexity: int = 50,
         iterations: int = 500,
         exaggeration: int = 6,
         metadata_protein: str = None,
         keep: bool = False,
         svg: bool = False,
         force: bool = False,
         show_cluster_numbers: bool = False
         ):
    """
    Run complete CASM analysis pipeline.

    Args:
        fasta (str): Path to input FASTA file
        output (str): Output basename
        subset (str, optional): Path to subset FASTA file. If None, creates random subset
        subset_size (int): Size of subset for alignment
        db_path (str): Path to GlobDB SQLite database
        threads (int): Number of threads
        perplexity (int): t-SNE perplexity parameter
        iterations (int): Number of optimization iterations
        exaggeration (int): Early exaggeration parameter
        metadata_protein (str): Path to protein metadata file
        keep (bool): Keep intermediate files
        svg (bool): Generate plot in SVG format
        force (bool): Force overwrite existing files
        show_cluster_numbers (bool): Display cluster enumeration in cluster centers in output plot

    Returns:
        sum_dict: Dictionary containing paths to all generated files
    """
    if (metadata_protein is not None and metadata_protein not in BASE_COLUMNS[1:] + ANNOTATION_COLUMNS +
            TAXONOMY_COLUMNS + CULTURE_COLLECTION_COLUMNS + HIGH_LEVEL_ENV_COLUMNS + LOW_LEVEL_ENV_COLUMNS):
        logger.error('Invalid metadata category. Please run "aastk metadata_categories" to display available options.')
        raise ValueError(f'Invalid metadata category: {metadata_protein}')


    logger.info("=== Starting Complete CASM Analysis ===")
    logger.info(f"Input FASTA: {fasta}")
    logger.info(f"Output basename: {output}")
    logger.info(f"Subset size: {subset_size}")
    logger.info(f"Threads: {threads}")

    intermediate_results = {}
    results = {}

    # Phase 1: Matrix construction
    logger.info("=== Phase 1: Matrix Construction ===")
    matrix_file, metadata_file = matrix(fasta=fasta,
                                        output=output,
                                        subset=subset,
                                        subset_size=subset_size,
                                        threads=threads,
                                        force=force
                                        )
    intermediate_results['matrix_file'] = matrix_file
    intermediate_results['metadata_file'] = metadata_file

    # Phase 2: t-SNE embedding
    logger.info("=== Phase 2: t-SNE Embedding ===")
    early_filename, final_filename = cluster(
        matrix_path=matrix_file,
        matrix_metadata_path=metadata_file,
        output=output,
        perplexity=perplexity,
        iterations=iterations,
        exaggeration=exaggeration,
        threads=threads,
        force=force
    )
    results['early_filename'] = early_filename
    results['final_filename'] = final_filename

    # Phase 3: Plot generation
    logger.info("=== Phase 3: Plot Generation ===")
    early_cluster_plot = casm_plot(
        clust_path=early_filename,
        output=output,
        metadata_protein=metadata_protein,
        db_path=db_path,
        svg=svg,
        force=force,
        show_cluster_numbers=show_cluster_numbers
    )

    full_cluster_plot = casm_plot(
        clust_path=early_filename,
        output=output,
        metadata_protein=metadata_protein,
        db_path=db_path,
        svg=svg,
        force=force,
        show_cluster_numbers=show_cluster_numbers
    )
    results['early_cluster_plot'] = early_cluster_plot
    results['full_cluster_plot'] = full_cluster_plot

    logger.info("=== Complete CASM Analysis Completed ===")

    if not keep and intermediate_results:
        logger.info("Cleaning up intermediate files")
        for key, filepath in intermediate_results.items():
            try:
                file_path = Path(filepath)
                if file_path.exists():
                    file_path.unlink()
                    logger.debug(f"Deleted {filepath}")
            except Exception as e:
                logger.warning(f"Failed to delete {filepath}: {e}")

    return





__version__ = '0.1.0'
__author__ = 'Marcel Rennig and Daan Speth'
__author_email__ = ['daan.speth@univie.ac.at', 'marcel.rennig@univie.ac.at']
__copyright__ = 'Copyright 2024'
__credits__ = ['Daan Speth', 'Marcel Rennig']
__description__ = 'toolkit for working with amino acid sequences obtained from GlobDB genomes'
__license__ = 'GPL3'
__maintainer__ = 'Marcel Rennig and Daan Speth'
__maintainer_email__ = 'daan.speth@univie.ac.at'
__name__ = 'aastk'
__python_requires__ = '>=3.13'
__status__ = 'Production/Stable'
__title__ = 'aastk'
__url__ = 'https://github.com/dspeth/aastk'
